import os
import numpy as np

from kgcnn.data.base import MemoryGraphDataset
from kgcnn.data.download import DownloadDataset


class CoraLuDataset(DownloadDataset, MemoryGraphDataset):
    r"""Store and process Cora dataset after `Lu et al. 2003 <https://www.aaai.org/Papers/ICML/2003/ICML03-066.pdf>`_ .

    Information in `Papers with code <https://paperswithcode.com/dataset/cora>`_ read:
    The Cora dataset consists of 2708 scientific publications classified into one of seven classes.
    The citation network consists of 5429 links. Each publication in the dataset is described by a 0/1-valued word
    vector indicating the absence/presence of the corresponding word from the dictionary.
    The dictionary consists of 1433 unique words.

    Downloaded from: `<https://linqs-data.soe.ucsc.edu/public/lbc/>`_ .

    References:

        (1) McCallum, A.K., Nigam, K., Rennie, J. et al. Automating the Construction of Internet Portals with Machine
            Learning. Information Retrieval 3, 127–163 (2000). https://doi.org/10.1023/A:1009953814988
        (2) Lu, Qing and Lise Getoor. “Link-Based Classification.” Encyclopedia of Machine Learning and Data
            Mining (2003).

    """

    download_info = {
        "dataset_name": "cora_lu",
        "data_directory_name": "cora_lu",
        "download_url": "https://linqs-data.soe.ucsc.edu/public/lbc/cora.tgz",
        # download_url = "https://linqs-data.soe.ucsc.edu/public/arxiv-mrdm05/arxiv.tar.gz"
        "download_file_name": 'cora.tgz',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": "cora_lu"
    }

    # Make cora graph that was published by Qing Lu, and Lise Getoor. "Link-based classification." ICML, 2003.
    # https://www.aaai.org/Papers/ICML/2003/ICML03-066.pdf

    def __init__(self, reload=False, verbose: int = 10):
        """Initialize Cora dataset after Lu et al. 2003.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        self.class_label_mapping = None
        # Use default base class init.

        MemoryGraphDataset.__init__(self, dataset_name="cora_lu", verbose=verbose)
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        if self.fits_in_memory:
            self.read_in_memory()

    def read_in_memory(self):
        """Load Cora data into memory and already split into items."""
        filepath = os.path.join(self.data_main_dir, self.data_directory_name, self.unpack_directory_name, "cora")

        ids = np.loadtxt(os.path.join(filepath, "cora.cites"))
        ids = np.array(ids, np.int)
        open_file = open(os.path.join(filepath, "cora.content"), "r")
        lines = open_file.readlines()
        labels = [x.strip().split('\t')[-1] for x in lines]
        nodes = [x.strip().split('\t')[0:-1] for x in lines]
        nodes = np.array([[int(y) for y in x] for x in nodes], dtype=np.int)
        open_file.close()
        # Match edge_indices not with ids but with edge_indices in nodes
        node_map = np.zeros(np.max(nodes[:, 0]) + 1, dtype=np.int)
        idx_new = np.arange(len(nodes))
        node_map[nodes[:, 0]] = idx_new
        indexlist = node_map[ids]
        order1 = np.argsort(indexlist[:, 1], axis=0, kind='mergesort')  # stable!
        ind1 = indexlist[order1]
        order2 = np.argsort(ind1[:, 0], axis=0, kind='mergesort')
        indices = ind1[order2]
        # Class mappings
        class_label_mapping = {'Genetic_Algorithms': 0,
                               'Reinforcement_Learning': 1,
                               'Theory': 2,
                               'Rule_Learning': 3,
                               'Case_Based': 4,
                               'Probabilistic_Methods': 5,
                               'Neural_Networks': 6}
        self.class_label_mapping = class_label_mapping

        label_id = np.array([class_label_mapping[x] for x in labels], dtype=np.int)
        labels = np.expand_dims(label_id, axis=-1)
        labels = np.array(labels == np.arange(7), dtype=np.float)

        self.assign_property("node_attributes", [nodes[:, 1:]])
        self.assign_property("edge_indices", [indices])
        self.assign_property("edge_attributes", [np.ones_like(indices)[:, :1]])
        self.assign_property("node_labels", [labels])
        self.assign_property("node_number", [label_id])
        self.assign_property("edge_weights", [np.ones_like(indices)[:, :1]])
        # self.graph_adjacency = make_adjacency_from_edge_indices(indices, np.ones_like(indices)[:, 0])

        return self

# ds = CoraLuDataset()



import os
import numpy as np

from kgcnn.data.datasets.GraphTUDataset2020 import GraphTUDataset2020


class MutagenicityDataset(GraphTUDataset2020):
    r"""Store and process Mutagenicity dataset from `TUDatasets <https://chrsmrrs.github.io/datasets/>`__ .

    Mutagenicity is a chemical compound dataset of drugs, which can be categorized into two classes:
    mutagen and non-mutagen.

    References:

        (1) Riesen, K. and Bunke, H.: IAM Graph Database Repository for Graph Based Pattern Recognition and
            Machine Learning. In: da Vitora Lobo, N. et al. (Eds.), SSPR&SPR 2008, LNCS, vol. 5342, pp. 287-297, 2008.

    """

    def __init__(self, reload=False, verbose: int = 10):
        """Initialize Mutagenicity dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        super(MutagenicityDataset, self).__init__("Mutagenicity", reload=reload, verbose=verbose)

    def read_in_memory(self):
        r"""Load Mutagenicity Dataset into memory and already split into items with further cleaning and
        processing.
        """
        super(MutagenicityDataset, self).read_in_memory()

        node_translate = np.array([6, 8, 17, 1, 7, 9, 35, 16, 15, 53, 11, 19, 3, 20], dtype="int")
        atoms_translate = ['C', 'O', 'Cl', 'H', 'N', 'F', 'Br', 'S', 'P', 'I', 'Na', 'ksb', 'Li', 'Ca']
        z_translate = {node_translate[i]: atoms_translate[i] for i in range(len(node_translate))}

        edge_indices = self.obtain_property("edge_indices")
        node_labels = self.obtain_property("node_labels")
        edge_labels = self.obtain_property("edge_labels")
        graph_labels = self.obtain_property("graph_labels")

        nodes = [node_translate[np.array(x, dtype="int")][:, 0] for x in node_labels]
        atoms = [[atoms_translate[int(y[0])] for y in x] for x in node_labels]
        edges = [x[:, 0]+1 for x in edge_labels]
        labels = graph_labels

        # Require cleaning steps
        labels_clean = []
        nodes_clean = []
        edge_indices_clean = []
        edges_clean = []
        atoms_clean = []

        # Remove unconnected atoms. not Na Li etc.
        self.info("Checking database...")
        for i in range(len(nodes)):
            nats = nodes[i]
            cons = np.arange(len(nodes[i]))
            test_cons = np.sort(np.unique(edge_indices[i].flatten()))
            is_cons = np.zeros_like(cons, dtype="bool")
            is_cons[test_cons] = True
            is_cons[nats == 20] = True  # Allow to be unconnected
            is_cons[nats == 3] = True  # Allow to be unconnected
            is_cons[nats == 19] = True  # Allow to be unconnected
            is_cons[nats == 11] = True  # Allow to be unconnected
            if np.sum(is_cons) != len(cons):
                info_list = nodes[i][is_cons == False]
                info_list, info_cnt = np.unique(info_list, return_counts=True)
                info_list = {z_translate[info_list[j]]: info_cnt[j] for j in range(len(info_list))}
                self.info("Removing unconnected %s from molecule %s" % (info_list, i))
                nodes_clean.append(nats[is_cons])
                atoms_clean.append([atoms[i][j] for j in range(len(is_cons)) if is_cons[j] == True])
                # Need to correct edge_indices
                indices_used = cons[is_cons]
                indices_new = np.arange(len(indices_used))
                indices_old = np.zeros(len(nodes[i]), dtype="int")
                indices_old[indices_used] = indices_new
                edge_idx_new = indices_old[edge_indices[i]]
                edge_indices_clean.append(edge_idx_new)
            else:
                nodes_clean.append(nats)
                atoms_clean.append(atoms[i])
                edge_indices_clean.append(edge_indices[i])
            edges_clean.append(edges[i])
            labels_clean.append(labels[i])

        self.info("Database still has unconnected Na+, Li+, ksb+ etc.")

        # Since no attributes in graph dataset, we use labels as attributes
        self.assign_property("graph_labels", labels_clean)
        self.assign_property("edge_indices", edge_indices_clean)
        self.assign_property("node_attributes", nodes_clean)
        self.assign_property("edge_attributes", edges_clean)
        self.assign_property("node_labels", nodes_clean)
        self.assign_property("edge_labels", edges_clean)
        self.assign_property("node_symbol", atoms_clean)
        self.assign_property("node_number", nodes_clean)
        self.assign_property("graph_size", [len(x) for x in nodes_clean])

        # return labels,nodes,edge_indices,edges,atoms
        return self



import unittest

import os
import random
import itertools
import tempfile

import numpy as np
import tensorflow as tf
import matplotlib as mpl
import matplotlib.pyplot as plt

from kgcnn.literature.MEGAN import MEGAN
from kgcnn.literature.MEGAN import shifted_sigmoid

from .utils import ASSETS_PATH

ks = tf.keras
# mpl.use('TkAgg')
mpl.use('Agg')


class TestMegan(unittest.TestCase):

    def random_input(self, num_batches, num_features):
        n = []
        ei = []
        e = []
        for b in range(num_batches):
            N = random.randint(5, 30)
            node_indices = list(range(N))
            M = random.randint(1, N - 1)

            n.append([[random.random() for _ in range(num_features)] for _ in range(N)])
            e.append([[random.random() for _ in range(num_features)] for _ in range(M)])
            ei.append(list(zip(random.sample(node_indices, M), random.sample(node_indices, M))))

        return (
            tf.ragged.constant(n, ragged_rank=1, dtype=tf.float32),
            tf.ragged.constant(e, ragged_rank=1, dtype=tf.float32),
            tf.ragged.constant(ei, ragged_rank=1, dtype=tf.int32),
        )

    # -- UNITTESTS --

    def test_plot_shifted_sigmoid(self):

        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10, 10))
        xs = np.linspace(0, 20, 1000)

        ax.set_title('Shifted sigmoid for MEGAN')
        ax.plot(xs, shifted_sigmoid(xs, multiplier=1), label='multiplier: 1')
        ax.plot(xs, shifted_sigmoid(xs, multiplier=2), label='multiplier: 2')
        ax.plot(xs, shifted_sigmoid(xs, multiplier=3), label='multiplier: 3')
        ax.legend()

        pdf_path = os.path.join(ASSETS_PATH, 'megan_shifted_sigmoid.pdf')
        fig.savefig(pdf_path)

    def test_shifted_sigmoid_setting_rules(self):
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10, 10))
        xs = np.linspace(0, 40, 1000)

        ax.set_title('Shifted sigmoid for MEGAN')
        ax.plot(xs, shifted_sigmoid(xs, shift=5, multiplier=1), label='shift: 5')
        ax.plot(xs, shifted_sigmoid(xs, shift=10, multiplier=2), label='shift: 10')
        ax.plot(xs, shifted_sigmoid(xs, shift=15, multiplier=3), label='shift: 15')
        ax.legend()

        pdf_path = os.path.join(ASSETS_PATH, 'megan_shifted_sigmoid_setting_rules.pdf')
        fig.savefig(pdf_path)

    def test_construction_basically_works(self):
        model = MEGAN(units=[1])
        self.assertIsInstance(model, MEGAN)
        self.assertIsInstance(model, ks.models.Model)

    def test_shapes_basically_work(self):
        num_batches = 5
        num_features = 3
        n, e, ei = self.random_input(num_batches=num_batches, num_features=num_features)

        # We just test a bunch of different configurations for the output dimension and the number of
        # explanation channels.
        output_dimensions = [1, 3]
        channel_dimensions = [2, 6]

        for num_out, num_channels in itertools.product(output_dimensions, channel_dimensions):
            model = MEGAN(
                units=[3],
                importance_channels=num_channels,
                final_units=[num_out],
            )

            out, node_importances, edge_importances = model([n, e, ei])
            self.assertEqual((num_batches, num_out), out.shape)
            self.assertEqual((num_batches, None, num_channels), node_importances.shape)
            self.assertEqual((num_batches, None, num_channels), node_importances.shape)

    def test_saving_loading_basically_works(self):
        num_batches = 5
        num_features = 3
        n, e, eid = self.random_input(num_batches=num_batches, num_features=num_features)

        num_channels = 2
        num_out = 1
        model = MEGAN(
            units=[5, 3],
            importance_channels=num_channels,
            final_units=[num_out]
        )
        # At this point the model is not built yet and should raise a value error
        with self.assertRaises(ValueError):
            model.summary()

        out, ni, ei = model([n, e, eid], training=False)
        self.assertEqual((num_batches, num_out), out.shape)
        self.assertEqual((num_batches, None, num_channels), ni.shape)
        self.assertEqual((num_batches, None, num_channels), ei.shape)

        # After having passed in some input, the model should be built and that should work
        model.summary()
        weights = model.get_weights()

        with tempfile.TemporaryDirectory() as path:
            model_path = os.path.join(path, 'model')

            # Saving the model to a file
            model.save(model_path)
            self.assertTrue(os.path.exists(model_path))

            # Loading the model from that file again
            model_loaded = ks.models.load_model(
                model_path,
                custom_objects={'MEGAN': MEGAN}
            )
            self.assertIsInstance(model_loaded, ks.models.Model)

            out_l, ni_l, ei_l = model_loaded([n, e, eid], training=False)
            # Basic test for correct shape
            self.assertEqual((num_batches, num_out), out_l.shape)
            self.assertEqual((num_batches, None, num_channels), ni_l.shape)
            self.assertEqual((num_batches, None, num_channels), ei_l.shape)

            # Comparing the model weights
            weights_l = model_loaded.get_weights()
            for w, w_l in zip(weights, weights_l):
                np.testing.assert_allclose(w, w_l)

            # Testing if both models produce the same outputs, given the same inputs
            np.testing.assert_allclose(out, out_l, rtol=1e-3)
            np.testing.assert_allclose(ni.values, ni_l.values, rtol=1e-3)
            np.testing.assert_allclose(ei.values, ei_l.values, rtol=1e-3)

    def test_explanation_only_training_basically_works(self):
        num_batches = 20
        num_features = 3
        n, e, eid = self.random_input(num_batches=num_batches, num_features=num_features)
        out = tf.ragged.constant([random.random() for _ in range(num_batches)])

        num_channels = 2
        num_out = 1
        model = MEGAN(
            units=[5, 3],
            importance_channels=num_channels,
            final_units=[num_out],
            return_importances=False,
            # We explicitly want to try using the explanation step here
            importance_factor=1.0,
            # We need this to indicate that we are trying to do regression here:
            regression_limits=(-1, 1),
            regression_reference=0
        )

        model.compile(
            optimizer='adam',
            loss=ks.losses.mean_squared_error
        )
        history = model.fit(
            [n, e, eid], out,
            batch_size=1,
            epochs=2,
            verbose=0
        )
        self.assertIn('exp_loss', history.history)
        self.assertNotEqual(0.0, history.history['exp_loss'])




import numpy as np

from kgcnn.data.datasets.GraphTUDataset2020 import GraphTUDataset2020


class MUTAGDataset(GraphTUDataset2020):
    r"""Store and process MUTAG dataset from `TUDatasets <https://chrsmrrs.github.io/datasets/>`__ .

    In `Papers with Code <https://paperswithcode.com/dataset/mutag>`__ :
    In particular, MUTAG is a collection of nitroaromatic compounds and the goal is to predict their mutagenicity
    on Salmonella typhimurium. Input graphs are used to represent chemical compounds, where vertices stand for atoms
    and are labeled by the atom type (represented by one-hot encoding), while edges between vertices represent
    bonds between the corresponding atoms. It includes 188 samples of chemical compounds with 7 discrete node labels.

    References:

        (1) Debnath, A.K., Lopez de Compadre, R.L., Debnath, G., Shusterman, A.J., and Hansch, C.
            Structure-activity relationship of mutagenic aromatic and heteroaromatic nitro compounds.
            J. Med. Chem. 34(2):786-797 (1991).
        (2) Nils Kriege, Petra Mutzel. 2012. Subgraph Matching Kernels for Attributed Graphs.
            International Conference on Machine Learning 2012.

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize MUTAG dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        super(MUTAGDataset, self).__init__("MUTAG", reload=reload, verbose=verbose)

    def read_in_memory(self, verbose: int = 1):
        r"""Load MUTAG data into memory and already split into items."""
        super(MUTAGDataset, self).read_in_memory()

        edge_labels = self.obtain_property("edge_labels")
        node_labels = self.obtain_property("node_labels")
        graph_labels = self.obtain_property("graph_labels")
        # split into separate graphs
        # graph_id, counts = np.unique(mutag_gi, return_counts=True)
        # graphlen = np.zeros(n_data, dtype=np.int)
        # graphlen[graph_id] = counts
        # nodes0123 = np.split(mutag_n, np.cumsum(graphlen)[:-1])
        node_translate = np.array([6, 7, 8, 9, 53, 17, 35], dtype="int")
        atoms_translate = ['C', 'N', 'O', 'F', 'I', 'Cl', 'Br']
        node_attributes = [node_translate[np.array(x, dtype="int")][:, 0] for x in node_labels]
        # nodes = [node_translate[x] for x in nodes0123]
        atoms = [[atoms_translate[int(y)] for y in x] for x in node_labels]
        graph_labels = np.array(graph_labels)
        graph_labels[graph_labels < 0] = 0

        self.assign_property("node_attributes", node_attributes)
        self.assign_property("edge_attributes", [x[:, 0] for x in edge_labels])
        self.assign_property("node_symbol", atoms)
        self.assign_property("node_number", node_attributes)
        self.assign_property("graph_labels", [x for x in graph_labels])
        self.assign_property("graph_size", [len(x) for x in node_attributes])

        return self

# data = MUTAGDataset()



import unittest

import numpy as np
import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.xai.testing import MockContext
from kgcnn.xai.base import MockImportanceExplanationMethod


# == UNIT TESTS ==

class TestMockImportanceExplanationMethod(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock = MockContext()
        cls.mock.__enter__()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock.__exit__(None, None, None)

    def test_basically_works(self):
        channels = 3
        xai_instance = MockImportanceExplanationMethod(channels=channels)
        node_importances, edge_importances = xai_instance(
            self.mock.model,
            self.mock.x,
            self.mock.y
        )
        assert isinstance(node_importances, tf.RaggedTensor)
        assert isinstance(edge_importances, tf.RaggedTensor)

        node_importances = node_importances.numpy()
        edge_importances = edge_importances.numpy()
        for v, w in zip(node_importances, edge_importances):
            assert v.shape[-1] == channels
            assert w.shape[-1] == channels



from kgcnn.data.transform.scaler.mol import *
import logging

# For old reference.
logging.error(
    "Module '%s' is deprecated and will be removed in future versions. Please move to 'kgcnn.data.transform.scaler'." % __name__)


import typing as t

import numpy as np


def flatten_importances_list(importances_list: t.List[np.ndarray],
                             ) -> t.List[float]:
    """
    Given an ``importances_list`` which is a list of numpy arrays (which may be different shapes), this
    function will flatten all the numpy arrays contained within the list and concatenate all the flattened
    arrays into a single long list of values.

    Args:
        importances_list: A list of numpy arrays which may have varying shapes.

    Returns:
        A list of all the values within the given arrays, flattened and concatenated in the order given.
    """
    flattened = []
    for importances in importances_list:
        flattened += importances.flatten().tolist()

    return flattened



import numpy as np
import re
import logging
import networkx as nx
from collections.abc import MutableMapping
from kgcnn.graph.serial import get_preprocessor
from typing import Any, Union
from copy import deepcopy, copy

logging.basicConfig()  # Module logger
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


# Base classes are collections.UserDict or collections.abc.MutableMapping. However, this comes with more
# code to replicate dict-behaviour. Here GraphDict inherits from dict. Checking for numpy arrays can be disabled
# in class variable. Set item is completely free.
class GraphDict(dict):
    r"""Dictionary container to store graph information in tensor-form.

    The tensors must be stored as numpy arrays. The naming convention is not restricted.
    The class is supposed to be handled just as a python dictionary.

    In addition, :obj:`assign_property` and :obj:`obtain_property` handles `None` values and cast into tensor format,
    when assigning a named value.

    Graph operations that modify edges or sort indices can be applied via :obj:`apply_preprocessor` located in
    :obj:`kgcnn.graph.preprocessor`.

    .. code-block:: python

        import numpy as np
        from kgcnn.graph.base import GraphDict
        graph = GraphDict({"edge_indices": np.array([[1, 0], [0, 1]]), "edge_labels": np.array([[-1], [1]])})
        graph.set("graph_labels", [0])  # opposite is get()
        graph.set("edge_attributes", [[1.0], [2.0]])
        graph.apply_preprocessor("add_edge_self_loops")
        graph.apply_preprocessor("sort_edge_indices")
        print(graph)

    """

    _require_str_key = True
    _cast_to_array = True
    _tensor_class = np.ndarray
    _tensor_conversion = np.array
    _require_validate = True

    def __init__(self, *args, **kwargs):
        r"""Initialize a new :obj:`GraphDict` instance."""
        super(GraphDict, self).__init__(*args, **kwargs)
        if self._require_validate:
            self.validate()
        
    def update(self, *args, **kwargs):
        super(GraphDict, self).update(*args, **kwargs)
        if self._require_validate:
            self.validate()

    def assign_property(self, key: str, value):
        r"""Add a named property as key, value pair to self. If the value is `None`, nothing is done.
        Similar to assign-item default method :obj:`__setitem__`, but ignores `None` values and casts to tensor.

        Args:
            key (str): Name of the graph tensor to add to self.
            value: Array or tensor value to add. Can also be None.

        Returns:
            None.
        """
        if self._require_str_key:
            if not isinstance(key, str):
                raise ValueError("`GraphDict` requires string keys, but kot '%s'." % key)
        if value is not None:
            if self._cast_to_array:
                if not isinstance(value, self._tensor_class):
                    value = self._tensor_conversion(value)
            self[key] = value

    def obtain_property(self, key: str):
        """Get tensor item by name. If key is not found, `None` is returned.

        Args:
            key (str): Name of the key to get value for.

        Returns:
            self[key].
        """
        if key in self:
            return self[key]
        return None

    def validate(self):
        """Routine to check if items are set correctly, i.e. string key and np.ndarray values.
        Could be done manually."""
        for key, value in self.items():
            if self._require_str_key:
                if not isinstance(key, str):
                    raise ValueError("`GraphDict` requires string keys, but kot '%s'." % key)
            if self._cast_to_array:
                if not isinstance(value, self._tensor_class):
                    self[key] = self._tensor_conversion(value)

    # Alias of internal assign_property and obtain property.
    set = assign_property
    # get = obtain_property

    def copy(self):
        return GraphDict({key: copy(value) for key, value in self.items()})

    def to_dict(self) -> dict:
        """Returns a python-dictionary of self. Does not copy values.

        Returns:
            dict: Dictionary of graph tensor objects.
        """
        return {key: value for key, value in self.items()}

    def search_properties(self, keys: Union[str, list]) -> list:
        r"""Search for properties in self.

        This includes a list of possible names or a pattern-matching of a single string.

        Args:
            keys (str, list): Pattern matching string or list of strings to search for.

        Returns:
            list: List of names in self that match :obj:`keys`.
        """
        if keys is None:
            return []
        elif isinstance(keys, str):
            match_props = []
            for x in self:
                if re.match(keys, x):
                    if re.match(keys, x).group() == x:
                        match_props.append(x)
            return sorted(match_props)
        elif isinstance(keys, (list, tuple)):
            # No pattern matching for list input.
            return sorted([x for x in keys if x in self])
        return []

    # Old Alias
    find_graph_properties = search_properties

    def assert_has_valid_key(self, key: str, raise_error: bool = True) -> bool:
        """Assert the property is found in self.

        Args:
            key (str): Name of property that must be defined.
            raise_error (bool): Whether to raise error. Default is True.

        Returns:
            bool: Key is valid. Only if `raise_error` is False.
        """
        if key not in self or self.obtain_property(key) is None:
            if raise_error:
                raise AssertionError("`GraphDict` does not have '%s' key." % key)
            return False
        return True

    def has_valid_key(self, key: str) -> bool:
        """Check if the property is found in self and also is not Noe.

        Args:
            key (str): Name of property that must be defined.

        Returns:
            bool: Key is valid. Only if `raise_error` is False.
        """
        if key not in self or self.obtain_property(key) is None:
            return False
        return True

    def from_networkx(self, graph: nx.Graph,
                      node_number: str = "node_number",
                      edge_indices: str = "edge_indices",
                      node_attributes: str = None,
                      edge_attributes: str = None,
                      node_labels: str = None):
        r"""Convert a networkx graph instance into a dictionary of graph-tensors. The networkx graph is always converted
        into integer node labels. The former node IDs can be hold in :obj:`node_labels`. Furthermore, node or edge
        data can be cast into attributes via :obj:`node_attributes` and :obj:`edge_attributes`.

        Args:
            graph (nx.Graph): A networkx graph instance to convert.
            node_number (str): The name that the node numbers are assigned to. Default is "node_number".
            edge_indices (str): The name that the edge indices are assigned to. Default is "edge_indices".
            node_attributes (str, list): Name of node attributes to add from node data. Can also be a list of names.
                Default is None.
            edge_attributes (str, list): Name of edge attributes to add from edge data. Can also be a list of names.
                Default is None.
            node_labels (str): Name of the labels of nodes to store former node IDs into. Default is None.

        Returns:
            self.
        """
        assert node_labels is None or isinstance(node_labels, str), "Please provide name of node labels or `None`"
        graph_int = nx.convert_node_labels_to_integers(graph, label_attribute=node_labels)
        graph_size = len(graph_int)

        def _attr_to_list(attr):
            if attr is None:
                attr = []
            elif isinstance(attr, str):
                attr = [attr]
            if not isinstance(attr, list):
                raise TypeError("Attribute name is neither list nor string.")
            return attr

        # Loop over nodes in graph.
        node_attr = _attr_to_list(node_attributes)
        if node_labels is not None:
            node_attr += [node_labels]
        node_attr_dict = {x: [None] * graph_size for x in node_attr}
        nodes_id = []
        for i, x in enumerate(graph_int.nodes.data()):
            nodes_id.append(x[0])
            for d in node_attr:
                if d not in x[1]:
                    raise KeyError("Node does not have property %s" % d)
                node_attr_dict[d][i] = x[1][d]

        edge_id = []
        edges_attr = _attr_to_list(edge_attributes)
        edges_attr_dict = {x: [None] * graph.number_of_edges() for x in edges_attr}
        for i, x in enumerate(graph_int.edges.data()):
            edge_id.append(x[:2])
            for d in edges_attr:
                if d not in x[2]:
                    raise KeyError("Edge does not have property %s" % d)
                edges_attr_dict[d][i] = x[2][d]

        # Storing graph tensors in self.
        self.assign_property(node_number, nodes_id)
        self.assign_property(edge_indices, edge_id)
        for key, value in node_attr_dict.items():
            self.assign_property(key, value)
        for key, value in edges_attr_dict.items():
            self.assign_property(key, value)
        return self

    def to_networkx(self, edge_indices: str = "edge_indices"):
        """Function draft to make a networkx graph. No attributes or data is supported at the moment.

        Args:
            edge_indices (str): Name of edge index tensors to make graph with. Default is "edge_indices".

        Returns:
            nx.DiGraph: Directed networkx graph instance.
        """
        graph = nx.DiGraph()
        graph.add_edges_from(self.obtain_property(edge_indices))
        return graph

    def apply_preprocessor(self, name, **kwargs):
        r"""Apply a preprocessor on self.

        Args:
            name: Name of a preprocessor that uses :obj:`kgcnn.graph.serial.get_preprocessor` for backward
                compatibility, or a proper serialization dictionary or class :obj:`GraphPreProcessorBase`.
            kwargs: Optional kwargs for preprocessor. Only used in connection with
                :obj:`kgcnn.graph.serial.get_preprocessor` if :obj:`name` is string.

        Returns:
            self
        """
        if isinstance(name, str):
            proc_graph = get_preprocessor(name, **kwargs)(self)
            # print(proc_graph)
            self.update(proc_graph)
            return self
        elif isinstance(name, GraphPreProcessorBase):
            proc_graph = name(self)
            self.update(proc_graph)
            return self
        raise ValueError("Unsupported preprocessor '%s'." % name)


# Deprecated alias for GraphDict.
GraphNumpyContainer = GraphDict


class GraphProcessorBase:
    r"""General base class for graph processors.

    A graph processor should be a subclass that implements a call methods and store all relevant information on how
    to change the graph properties and how they are supposed to be named in the :obj:`GraphDict` .

    .. code-block:: python

        from kgcnn.graph.base import GraphDict, GraphProcessorBase

        class NewProcessor(GraphProcessorBase):

            def __init__(self, name="new_processor", node_property_name="node_number"):
                self.node_property_name = node_property_name
                self.name = name

            def __call__(self, graph: GraphDict) -> GraphDict:
                # Do nothing but return the same property.
                return GraphDict({self.node_property_name: graph[self.node_property_name]})

            def get_config(self):
                return {"name": self.name, "node_property_name": self.node_property_name}

    The class provides utility functions to assign and obtain graph properties to or from a :obj:`GraphDict` by name.
    Since the namespace on how to name properties in a general :obj:`GraphDict` is not restricted, subclasses of
    :obj:`GraphProcessorBase` must set the property names in class construction and then can use the utility functions
    :obj:`_obtain_properties` and :obj:`_assign_properties`.

    Additionally, :obj:`_obtain_properties` and :obj:`_assign_properties` can contain a flexible list of properties
    or a search string that will return multiple properties. This is mainly required to extract a flexible list of
    properties, for example all edge-related properties via the search string '^edge_(?!indices$).*' that starts with
    'edge_*' but not 'edge_indices'. Note that you can always cast to a networkx graph, operate your own functions,
    then transform back into tensor form and wrap it as a :obj:`GraphProcessorBase` .
    """

    @staticmethod
    def _obtain_properties(graph: GraphDict, to_obtain: dict, to_search, to_silent) -> dict:
        r"""Extract a dictionary of named properties from a :obj:`GraphDict`.

        Args:
            graph (GraphDict): Dictionary with of graph properties.
            to_obtain (dict): Dictionary of keys that holds names or a list of names or a search string of graph
                properties to fetch. The output dictionary will then have the same keys but with arrays in place of the
                property name(s). The keys can be used as function arguments for some transform function.
            to_search (list): A list of strings that should be considered search strings.
            to_silent (list): A list of strings that suppress 'error not found' messages.

        Returns:
            dict: A dictionary of resolved graph properties in tensor form.
        """
        obtained_properties = {}
        for key, name in to_obtain.items():
            if isinstance(name, str):
                if name in to_search:
                    names = graph.search_properties(name)  # Will be sorted list and existing only
                    obtained_properties[key] = [graph.obtain_property(x) for x in names]
                else:
                    if not graph.has_valid_key(name) and key not in to_silent:
                        module_logger.warning("Missing '%s' in '%s'" % (name, type(graph).__name__))
                    obtained_properties[key] = graph.obtain_property(name)
            elif isinstance(name, (list, tuple)):
                prop_list = []
                for x in name:
                    if not graph.has_valid_key(x) and key not in to_silent:
                        module_logger.warning("Missing '%s' in '%s'" % (x, type(graph).__name__))
                    prop_list.append(graph.obtain_property(x))
                obtained_properties[key] = prop_list
            else:
                raise ValueError("Unsupported property identifier %s" % name)
        return obtained_properties

    @staticmethod
    def _assign_properties(graph: GraphDict, graph_properties: Union[list, np.ndarray],
                           to_assign, to_search, in_place=False) -> Union[dict, GraphDict]:
        r"""Assign a list of arrays to a :obj:`GraphDict` by name.

        Args:
            graph (GraphDict): Dictionary with of graph properties. That can be changed in place. Also required to
                resolve search strings.
            graph_properties (list, np.ndarray): Array or list of arrays which are new graph properties
            to_assign (str, list): Name-keys or list of name-keys for the :obj:`graph_properties`. If :obj:`to_assign`
                is list, the length must match :obj:`graph_properties` .
            to_search (list): A list of strings that should be considered search strings.
                Note that in this case there must be a list of arrays in place of a single array, since search
                strings will return a list of matching strings.
            in_place (bool): Whether to update :obj:`graph` argument.

        Returns:
            dict: Dictionary of arrays matched with name-keys.
        """
        out_graph = GraphDict()

        def _check_list_property(n, p):
            if not isinstance(p, (list, tuple)):
                module_logger.error("Wrong return type for '%s', which is not list." % n)
            if len(n) != len(p):
                module_logger.error("Wrong number of properties '%s' for '%s'." % (p, n))

        def _assign_single(name, single_graph_property):
            if isinstance(name, str):
                if name in to_search:
                    names = graph.search_properties(name)  # Will be sorted list and existing only
                    # Assume that names matches graph_properties
                    _check_list_property(names, single_graph_property)
                    for x, gp in zip(names, single_graph_property):
                        out_graph.assign_property(x, gp)
                else:
                    out_graph.assign_property(name, single_graph_property)
            elif isinstance(name, (list, tuple)):
                # For nested output
                _check_list_property(name, single_graph_property)
                for x, gp in zip(name, single_graph_property):
                    out_graph.assign_property(x, gp)
            else:
                module_logger.error("Wrong type of named property '%s'" % name)
            return

        # Process assignment here.
        if isinstance(to_assign, str):
            _assign_single(to_assign, graph_properties)
        elif isinstance(to_assign, (list, tuple)):
            _check_list_property(to_assign, graph_properties)
            for key, value in zip(to_assign, graph_properties):
                _assign_single(key, value)

        if in_place:
            graph.update(out_graph)
        return out_graph

    @staticmethod
    def has_special_characters(query, pat=re.compile("[@\\\\!#$%^&*()<>?/|}{~:]")):
        """Whether a query string has special characters."""
        return pat.search(query) is not None


class GraphPreProcessorBase(GraphProcessorBase):
    r"""Base wrapper for a graph preprocessor to use general transformation functions.

    This class inherits from :obj:`GraphProcessorBase` and makes use of :obj:`_obtain_properties` and
    :obj:`_assign_properties` in a predefined :obj:`__call__` method. The base class :obj:`__call__` method retrieves
    graph properties by name, passes them as kwargs to the :obj:`call` method, which must be implemented by subclasses,
    and assigns the (array-like) output to a :obj:`GraphDict` by name. To specify the kwargs for call and the names that
    are expected to be set in the graph dictionary, the class attributes :obj:`self._to_obtain` ,
    :obj:`self._to_assign` , :obj:`self._to_assign` , :obj:`self._call_kwargs` , :obj:`self._search` must be set
    appropriately in the constructor. The design of this class should be similar to :obj:`ks.layers.Layer` .
    Example implementation for subclass:

    .. code-block:: python

        from kgcnn.graph.base import GraphDict, GraphPreProcessorBase

        class NewProcessor(GraphPreProcessorBase):

            def __init__(self, name="new_processor", node_property_name="node_number", **kwargs)
                super().__init__(name=name, **kwargs):
                self._to_obtain = {"nodes": node_property_name}  # what 'call' needs as properties
                self._to_assign = ["new_nodes"]  # will be the output of call()
                self._call_kwargs = {"do_nothing": True}
                self._config_kwargs.update({"node_property_name": node_property_name})

            def call(self, nodes: np.ndarray, do_nothing: bool):
                # Do nothing but return the same property.
                return nodes

    For proper serialization you should update :obj:`self._config_kwargs` . For information on :obj:`self._search` and
    :obj:`self._silent` see docs of :obj:`GraphProcessorBase` .

    """

    def __init__(self, *, in_place: bool = False, name: str = None):
        r"""Initialize class.

        Args:
            in_place (bool): Whether to update graph dict. Default is False.
            name: Name of the preprocessor.
        """
        self._config_kwargs = {"in_place": in_place, "name": name}
        self._in_place = in_place
        # Below attributes must be set by subclass.
        self._to_obtain = {}
        self._to_assign = None
        self._call_kwargs = {}
        self._search = []
        self._silent = []

    def call(self, **kwargs):
        raise NotImplementedError("Must be implemented in sub-class.")

    def __call__(self, graph: GraphDict):
        r"""Processing the graph with function given in :obj:`call` .

        Args:
            graph (GraphDict): Graph dictionary to process.

        Returns:
            dict: Dictionary of new properties.
        """
        graph_properties = self._obtain_properties(graph, self._to_obtain, self._search, self._silent)
        # print(graph_properties)
        processed_properties = self.call(**graph_properties, **self._call_kwargs)
        out_graph = self._assign_properties(graph, processed_properties, self._to_assign, self._search)
        # print(out_graph)
        if self._in_place:
            graph.update(out_graph)
            return graph
        return out_graph

    def get_config(self):
        r"""Copy config from :obj:`self._config_kwargs` ."""
        config = deepcopy(self._config_kwargs)
        return config


class GraphPostProcessorBase(GraphProcessorBase):
    r"""Base wrapper for a graph postprocessor to use general transformation functions.

    This class inherits from :obj:`GraphProcessorBase` and makes use of :obj:`_obtain_properties` and
    :obj:`_assign_properties` in a predefined :obj:`__call__` method. The base class :obj:`__call__` method retrieves
    graph properties by name, passes them as kwargs to the :obj:`call` method, which must be implemented by subclasses,
    and assigns the (array-like) output to a :obj:`GraphDict` by name. To specify the kwargs for call and the names that
    are expected to be set in the graph dictionary, the class attributes :obj:`self._to_obtain` ,
    :obj:`self._to_assign` , :obj:`self._to_assign` , :obj:`self._call_kwargs` , :obj:`self._search` must be set
    appropriately in the constructor. The design of this class should be similar to :obj:`ks.layers.Layer` .
    Example implementation for subclass:

    .. code-block:: python

        from kgcnn.graph.base import GraphDict, GraphPostProcessorBase

        class NewProcessor(GraphPostProcessorBase):

            def __init__(self, name="new_processor", node_property_name="node_number", **kwargs)
                super().__init__(name=name, **kwargs):
                self._to_obtain = {"nodes": node_property_name}  # what 'call' needs as properties
                self._to_obtain_pre = {}
                self._to_assign = ["new_nodes"]  # will be the output of call()
                self._call_kwargs = {"do_nothing": True}
                self._config_kwargs.update({"node_property_name": node_property_name})

            def call(self, nodes: np.ndarray, do_nothing: bool):
                # Do nothing but return the same property.
                return nodes

    For proper serialization you should update :obj:`self._config_kwargs` . For information on :obj:`self._search` and
    :obj:`self._silent` see docs of :obj:`GraphProcessorBase` .

    Note that for :obj:`GraphPostProcessorBase` one can also set :obj:`self._to_obtain_pre` if information of the input
    graph or a previous graph is required, which is common for postprocessing.
    """

    def __init__(self, *, in_place: bool = False, name: str = None):
        r"""Initialize class.

        Args:
            in_place (bool): Whether to update graph dict. Default is False.
            name: Name of the preprocessor.
        """
        self._config_kwargs = {"in_place": in_place, "name": name}
        self._in_place = in_place
        # Below attributes must be set by subclass.
        self._to_obtain = {}
        self._to_obtain_pre = {}
        self._to_assign = None
        self._call_kwargs = {}
        self._search = []
        self._silent = []

    def call(self, **kwargs):
        raise NotImplementedError("Must be implemented in sub-class.")

    def __call__(self, graph: GraphDict, pre_graph: GraphDict = None):
        r"""Processing the graph with function given in :obj:`call` .

        Args:
            graph (GraphDict): Graph dictionary to process.
            pre_graph (GraphDict): Additional graph dictionary with properties to include in :obj:`call` .
                This is the main difference to preprocessor. Defaults to None.

        Returns:
            dict: Dictionary of new properties.
        """
        graph_properties_y = self._obtain_properties(graph, self._to_obtain, self._search, self._silent)

        if pre_graph is not None:
            graph_properties_x = self._obtain_properties(pre_graph, self._to_obtain_pre, self._search, self._silent)
            processed_properties = self.call(**graph_properties_y, **graph_properties_x, **self._call_kwargs)
        else:
            processed_properties = self.call(**graph_properties_y, **self._call_kwargs)

        out_graph = self._assign_properties(graph, processed_properties, self._to_assign, self._search)
        if self._in_place:
            graph.update(out_graph)
            return graph
        return out_graph

    def get_config(self):
        r"""Copy config from :obj:`self._config_kwargs` ."""
        config = deepcopy(self._config_kwargs)
        return config



import os
import numpy as np

from kgcnn.data.datasets.GraphTUDataset2020 import GraphTUDataset2020
from kgcnn.mol.encoder import OneHotEncoder


class PROTEINSDataset(GraphTUDataset2020):
    r"""Store and process PROTEINS dataset from `TUDatasets <https://chrsmrrs.github.io/datasets/>`__ .

    In `Papers with Code <https://paperswithcode.com/dataset/proteins>`__ :
    PROTEINS is a dataset of proteins that are classified as enzymes or non-enzymes.
    Nodes represent the amino acids and two nodes are connected by an edge if they are less than 6 Angstroms apart.

    References:

        (1) K. M. Borgwardt, C. S. Ong, S. Schoenauer, S. V. N. Vishwanathan, A. J. Smola, and H. P. Kriegel.
            Protein function prediction via graph kernels. Bioinformatics, 21(Suppl 1):i47–i56, Jun 2005.
        (2) P. D. Dobson and A. J. Doig. Distinguishing enzyme structures from non-enzymes without alignments.
            J. Mol. Biol., 330(4):771–783, Jul 2003.

    """

    def __init__(self, reload=False, verbose: int = 10):
        """Initialize MUTAG dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        # We set dataset_name to None since all flags are defined by hand in subclass definition.
        super(PROTEINSDataset, self).__init__(dataset_name="PROTEINS", reload=reload, verbose=verbose)

    def read_in_memory(self):
        r"""Load PROTEINS Dataset into memory and already split into items with further cleaning and
        processing.
        """
        super(PROTEINSDataset, self).read_in_memory()
        # One-hot encoders
        ohe = OneHotEncoder(
            [-538, -345, -344, -134, -125, -96, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
             21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 41, 42, 47, 61, 63, 73, 74, 75,
             82, 104, 353, 355, 360, 558, 797, 798], add_unknown=False)
        ohe2 = OneHotEncoder([0, 1, 2], add_unknown=False)
        ohe3 = OneHotEncoder([i for i in range(0, 17)], add_unknown=False)

        graph_labels = self.obtain_property("graph_labels")
        node_attributes = self.obtain_property("node_attributes")
        node_labels = self.obtain_property("node_labels")
        node_degree = self.obtain_property("node_degree")
        self.assign_property("graph_labels", [x - 1 for x in graph_labels])
        self.assign_property("node_attributes", [np.array([ohe(int(y)) for y in x]) for x in node_attributes])
        self.assign_property("node_labels", [np.array([ohe2(int(y)) for y in x]) for x in node_labels])
        self.assign_property("node_degree", [np.array([ohe3(int(y)) for y in x]) for x in node_degree])
        self.assign_property("graph_size", [len(x) if x is not None else None for x in node_attributes])

        return self

# ds = PROTEINSDataset()



import tensorflow as tf
from typing import Union
from kgcnn.layers.casting import ChangeTensorType
from kgcnn.model.utils import get_model_class

ks = tf.keras


@ks.utils.register_keras_serializable(package='kgcnn', name='EnergyForceModel')
class EnergyForceModel(ks.models.Model):
    r"""Force model that generates forces from any energy predicting model by taking the derivative with respect to
    the input coordinates.

    For now the model has to cast to dense tensor for using :obj:`batch_jacobian` , however, this will likely support
    ragged tensors in the future.

    .. code-block:: python

        import tensorflow as tf
        from kgcnn.model.force import EnergyForceModel
        model = EnergyForceModel(
            model_energy= {
                "module_name": "kgcnn.literature.Schnet",
                "class_name": "make_model",
                "config": {
                    "name": "SchnetEnergy",
                    "inputs": [
                        {"shape": [None], "name": "z", "dtype": "float32", "ragged": True},
                        {"shape": [None, 3], "name": "R", "dtype": "float32", "ragged": True},
                        {"shape": [None, 2], "name": "range_indices", "dtype": "int64", "ragged": True}
                    ],
                    "input_embedding": {
                        "node": {"input_dim": 95, "output_dim": 128}
                    },
                    "last_mlp": {"use_bias": [True, True, True], "units": [128, 64, 1],
                                 "activation": ['kgcnn>shifted_softplus', 'kgcnn>shifted_softplus', 'linear']},
                    "interaction_args": {
                        "units": 128, "use_bias": True, "activation": "kgcnn>shifted_softplus", "cfconv_pool": "sum"
                    },
                    "node_pooling_args": {"pooling_method": "sum"},
                    "depth": 6,
                    "gauss_args": {"bins": 25, "distance": 5, "offset": 0.0, "sigma": 0.4}, "verbose": 10,
                    "output_embedding": "graph",
                    "use_output_mlp": False,
                    "output_mlp": None
                }
            },
            coordinate_input=1,
            output_as_dict=True,
            output_to_tensor=False,
            output_squeeze_states=True,
            is_physical_force=True
        )

    """

    def __init__(self,
                 model_energy=None,
                 coordinate_input: Union[int, str] = 1,
                 output_as_dict: bool = True,
                 ragged_validate: bool = False,
                 output_to_tensor: bool = True,
                 output_squeeze_states: bool = False,
                 nested_model_config: bool = True,
                 is_physical_force: bool = True,
                 **kwargs):
        r"""Initialize :obj:`EnergyForceModel` with sub-model for energy prediction.

        This wrapper model was designed for models in `kgcnn.literature` that predict energy from geometric
        information.

        .. note::

            The (serialized) energy model is inferred by `module_name` , `class_name` , `config` within :obj:`kgcnn` ,
            but you can also pass any model directly to `model_energy` and
            set `nested_model_config` to `False` .

        Args:
            model_energy (dict): Keras model for energy prediction. Can also be a serialization dict.
            coordinate_input (str, int): Index or key where to find coordinate tensor in model input.
            output_as_dict (bool): Whether to return energy and force as list or as dict. Default is True.
            ragged_validate (bool): Whether to validate ragged tensor creation. Default is False.
            output_to_tensor (bool): Whether to cast the output to tensor or keep ragged output. Default is True
            output_squeeze_states (bool): Whether to squeeze states, which can be done for one energy value to remove
                an axis of one.
            nested_model_config (bool): Whether `config` has model config of the energy model. Default is True.
            is_physical_force (bool): Whether gradient of force, which is the negative gradient, is to be returned.
        """
        super(EnergyForceModel, self).__init__(self, **kwargs)
        if model_energy is None:
            raise ValueError("Require valid model in `model_energy` for force prediction.")

        # Input for model_energy.
        self._model_energy = model_energy

        if isinstance(model_energy, ks.models.Model):
            # Ignoring module_name and class_name.
            self.energy_model = model_energy
        elif isinstance(model_energy, dict):
            if "module_name" not in model_energy:
                self.energy_model = ks.utils.deserialize_keras_object(model_energy)
            else:
                self.energy_model_class = get_model_class(model_energy["module_name"], model_energy["class_name"])
                self.energy_model = self.energy_model_class(**model_energy["config"])
        else:
            raise TypeError("Input `model_energy` must be dict or `ks.models.Model` .")

        # Additional parameters of io and behavior of this class.
        self.ragged_validate = ragged_validate
        self.coordinate_input = coordinate_input
        self.output_as_dict = output_as_dict
        self.output_to_tensor = output_to_tensor
        self.output_squeeze_states = output_squeeze_states
        self.is_physical_force = is_physical_force
        self.nested_model_config = nested_model_config

        # Layers.
        self.cast_coordinates = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="mask")

    def call(self, inputs, training=False, **kwargs):
        """Forward pass that wraps energy model in gradient tape.

        Args:
            inputs (list, dict): Must be list of (tensor) input for energy model.
                Index or key to find coordinates must be provided.
            training (bool): Whether model is in training, passed to energy model. Default is False.

        Returns:
            dict, list: Model output plus force or derivative.
        """
        x = inputs[self.coordinate_input]
        inputs_energy = [i for i in inputs]
        # x is ragged tensor of shape (batch, [N], 3) with cartesian coordinates.
        # `batch_jacobian` does not yet support ragged tensor input.
        # Cast to masked tensor for coordinates only.
        x_pad, x_mask = self.cast_coordinates(x, training=training, **kwargs)  # (batch, N, 3), (batch, N, 3)
        with tf.GradientTape() as tape:
            tape.watch(x_pad)
            # Temporary solution for casting.
            # Cast back to ragged tensor for model input.
            x_pad_to_ragged = self._cast_coordinates_pad_to_ragged(x_pad, x_mask, self.ragged_validate)
            inputs_energy[self.coordinate_input] = x_pad_to_ragged
            # Predict energy.
            # Energy must be tensor of shape (batch, states)
            eng = self.energy_model(inputs_energy, training=training, **kwargs)
        e_grad = tape.batch_jacobian(eng, x_pad)
        e_grad = tf.transpose(e_grad, perm=[0, 2, 3, 1])

        if self.is_physical_force:
            e_grad = -e_grad

        if self.output_squeeze_states:
            e_grad = tf.squeeze(e_grad, axis=-1)
        if not self.output_to_tensor:
            e_grad = self._cast_coordinates_pad_to_ragged(e_grad, x_mask, self.ragged_validate)
        if self.output_as_dict:
            return {"energy": eng, "force": e_grad}
        else:
            return eng, e_grad

    # Temporary solution.
    @staticmethod
    @tf.function
    def _cast_coordinates_pad_to_ragged(x_pad, x_mask, validate):
        # Assume that float mask is the same for all coordinates.
        x_mask_number = tf.cast(x_mask[:, :, 0], dtype="bool")  # (batch, N)
        x_values = x_pad[x_mask_number]
        x_row_length = tf.reduce_sum(tf.cast(x_mask_number, dtype="int64"), axis=-1)
        return tf.RaggedTensor.from_row_lengths(x_values, x_row_length, validate=validate)

    def get_config(self):
        """Get config."""
        # Keras model does not provide config from base class.
        # conf = super(EnergyForceModel, self).get_config()
        conf = {}
        # Serialize class if _model_energy is not dict.
        if isinstance(self._model_energy, dict):
            model_energy = self._model_energy
        else:
            model_energy = ks.utils.serialize_keras_object(self._model_energy)
        conf.update({
            "model_energy": model_energy,
            "coordinate_input": self.coordinate_input,
            "output_as_dict": self.output_as_dict,
            "ragged_validate": self.ragged_validate,
            "output_to_tensor": self.output_to_tensor,
            "output_squeeze_states": self.output_squeeze_states,
            "nested_model_config": self.nested_model_config
        })
        return conf



import os

from kgcnn.data.tudataset import GraphTUDataset
from kgcnn.data.download import DownloadDataset


class GraphTUDataset2020(GraphTUDataset, DownloadDataset):
    r"""Base class for loading graph datasets published by `TU Dortmund University
    <https://chrsmrrs.github.io/datasets>`_ .

    This general base class has functionality to load TUDatasets in a generic way.

    .. note::

        Note that subclasses of `GraphTUDataset` in :obj:``kgcnn.data.datasets`` should still be made,
        if the dataset needs more refined post-precessing. Not all datasets can provide all types of graph
        properties like `edge_attributes` etc.

    References:

        (1) TUDataset: A collection of benchmark datasets for learning with graphs.
            Christopher Morris and Nils M. Kriege and Franka Bause and Kristian Kersting and Petra Mutzel and
            Marion Neumann, ICML 2020 Workshop on Graph Representation Learning and Beyond (GRL+ 2020)
            `<www.graphlearning.io>`_ .

    """

    # List of datasets in TUDatasets.
    tudataset_ids = [
        # Molecules
        "AIDS", "alchemy_full", "aspirin", "benzene", "BZR", "BZR_MD", "COX2", "COX2_MD", "DHFR", "DHFR_MD", "ER_MD",
        "ethanol", "FRANKENSTEIN", "malonaldehyde", "MCF-7", "MCF-7H", "MOLT-4", "MOLT-4H", "Mutagenicity", "MUTAG",
        "naphthalene", "NCI1", "NCI109", "NCI-H23", "NCI-H23H", "OVCAR-8", "OVCAR-8H", "P388", "P388H", "PC-3", "PC-3H",
        "PTC_FM", "PTC_FR", "PTC_MM", "PTC_MR", "QM9", "salicylic_acid", "SF-295", "SF-295H", "SN12C", "SN12CH",
        "SW-620", "SW-620H", "toluene", "Tox21_AhR_training", "Tox21_AhR_testing", "Tox21_AhR_evaluation",
        "Tox21_AR_training", "Tox21_AR_testing", "Tox21_AR_evaluation", "Tox21_AR-LBD_training", "Tox21_AR-LBD_testing",
        "Tox21_AR-LBD_evaluation", "Tox21_ARE_training", "Tox21_ARE_testing", "Tox21_ARE_evaluation",
        "Tox21_aromatase_training", "Tox21_aromatase_testing", "Tox21_aromatase_evaluation", "Tox21_ATAD5_training",
        "Tox21_ATAD5_testing", "Tox21_ATAD5_evaluation", "Tox21_ER_training", "Tox21_ER_testing", "Tox21_ER_evaluation",
        "Tox21_ER-LBD_training", "Tox21_ER-LBD_testing", "Tox21_ER-LBD_evaluation", "Tox21_HSE_training",
        "Tox21_HSE_testing", "Tox21_HSE_evaluation", "Tox21_MMP_training", "Tox21_MMP_testing", "Tox21_MMP_evaluation",
        "Tox21_p53_training", "Tox21_p53_testing", "Tox21_p53_evaluation", "Tox21_PPAR-gamma_training",
        "Tox21_PPAR-gamma_testing", "Tox21_PPAR-gamma_evaluation", "UACC257", "UACC257H", "uracil", "Yeast", "YeastH",
        "ZINC_full", "ZINC_test", "ZINC_train", "ZINC_val",
        # Bioinformatics
        "DD", "ENZYMES", "KKI", "OHSU", "Peking_1", "PROTEINS", "PROTEINS_full",
        # Computer vision
        "COIL-DEL", "COIL-RAG", "Cuneiform", "Fingerprint", "FIRSTMM_DB", "Letter-high", "Letter-low", "Letter-med",
        "MSRC_9", "MSRC_21", "MSRC_21C",
        # Social networks
        "COLLAB", "dblp_ct1", "dblp_ct2", "DBLP_v1", "deezer_ego_nets", "facebook_ct1", "facebook_ct2",
        "github_stargazers", "highschool_ct1", "highschool_ct2", "IMDB-BINARY", "IMDB-MULTI", "infectious_ct1",
        "infectious_ct2", "mit_ct1", "mit_ct2", "REDDIT-BINARY", "REDDIT-MULTI-5K", "REDDIT-MULTI-12K",
        "reddit_threads", "tumblr_ct1", "tumblr_ct2", "twitch_egos", "TWITTER-Real-Graph-Partial",
        # Synthetic
        "COLORS-3", "SYNTHETIC", "SYNTHETICnew", "Synthie", "TRIANGLES"
    ]

    def __init__(self, dataset_name: str, reload: bool = False, verbose: int = 10):
        """Initialize a `GraphTUDataset` instance from string identifier.

        Args:
            dataset_name (str): Name of a dataset.
            reload (bool): Download the dataset again and prepare data on disk.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        if not isinstance(dataset_name, str):
            raise ValueError("Please provide string identifier for TUDataset.")

        GraphTUDataset.__init__(self, verbose=verbose, dataset_name=dataset_name)

        # Prepare download
        if dataset_name in self.tudataset_ids:
            self.download_info = {
                "data_directory_name": dataset_name,
                "download_url": "https://www.chrsmrrs.com/graphkerneldatasets/" + dataset_name + ".zip",
                "download_file_name": dataset_name + ".zip",
                "unpack_zip": True,
                "unpack_directory_name": dataset_name,
                "dataset_name": dataset_name
            }
        else:
            raise ValueError("Can not resolve %s as a TUDataset." % dataset_name,
                             "Add to `all_tudataset_identifier` list manually.")

        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name)
        self.file_directory = os.path.join(self.unpack_directory_name, dataset_name)
        self.fits_in_memory = True

        if self.fits_in_memory:
            self.read_in_memory()

    @staticmethod
    def _debug_read_list():
        line_ids = []
        with open("datasets.md", 'r') as f:
            for line in f.readlines():
                if line[:3] == "|**":
                    line_ids.append(line.split("**")[1])
        return line_ids



import os
import pickle
import numpy as np
import scipy.io
import json
import pandas as pd
from typing import Union
from kgcnn.data.qm import QMDataset
from kgcnn.data.download import DownloadDataset
from kgcnn.mol.io import write_list_to_xyz_file
from kgcnn.graph.geom import coulomb_matrix_to_inverse_distance_proton, coordinates_from_distance_matrix
from kgcnn.graph.adj import invert_distance
from kgcnn.mol.methods import inverse_global_proton_dict


class QM7Dataset(QMDataset, DownloadDataset):
    r"""Store and process QM7 dataset from `Quantum Machine <http://quantum-machine.org/datasets/>`__ . dataset.

    From `Quantum Machine <http://quantum-machine.org/datasets/>`__ :
    This dataset is a subset of GDB-13 (a database of nearly 1 billion stable and synthetically accessible
    organic molecules) composed of all molecules of up to 23 atoms (including 7 heavy atoms C, N, O, and S),
    totalling 7165 molecules. We provide the Coulomb matrix representation of these molecules and their atomization
    energies computed similarly to the FHI-AIMS implementation of the Perdew-Burke-Ernzerhof hybrid functional (PBE0).
    This dataset features a large variety of molecular structures such as double and triple bonds, cycles, carboxy,
    cyanide, amide, alcohol and epoxy.
    The atomization energies are given in kcal/mol and are ranging from -800 to -2000 kcal/mol.
    The dataset is composed of three multidimensional arrays X (7165 x 23 x 23), Tm(7165) and P (5 x 1433)
    representing the inputs (Coulomb matrices), the labels (atomization energies) and the splits for cross-validation,
    respectively. The dataset also contain two additional multidimensional arrays Z (7165) and R (7165 x 3)
    representing the atomic charge and the cartesian coordinate of each atom in the molecules.

    Here, the coordinates are given and converted with :obj:`QMDataset` to molecular structure.
    Labels are not scaled but have original units. Original splits are added to the dataset.

    References:

        (1) L. C. Blum, J.-L. Reymond, 970 Million Druglike Small Molecules for Virtual Screening in the Chemical
            Universe Database GDB-13, J. Am. Chem. Soc., 131:8732, 2009.
        (2) M. Rupp, A. Tkatchenko, K.-R. Müller, O. A. von Lilienfeld: Fast and Accurate Modeling of Molecular
            Atomization Energies with Machine Learning, Physical Review Letters, 108(5):058301, 2012.

    """

    download_info = {
        "dataset_name": "QM7",
        "data_directory_name": "qm7",
        # https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm7.mat
        "download_url": "http://quantum-machine.org/data/qm7.mat",
        "download_file_name": 'qm7.mat',
        "unpack_tar": False,
        "unpack_zip": False,
    }

    def __init__(self, reload: bool = False, verbose: int = 10):
        """Initialize QM9 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        QMDataset.__init__(self, verbose=verbose, dataset_name="QM7")
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.label_names = ["u0_atom"]
        self.label_units = ["kcal/mol"]
        self.label_unit_conversion = np.array([1.0] * 14)  # Pick always same units for training
        self.dataset_name = "QM7"
        self.require_prepare_data = True
        self.fits_in_memory = True
        self.verbose = verbose
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name)
        self.file_name = "qm7.csv"

        if self.require_prepare_data:
            self.prepare_data(overwrite=reload)

        if self.fits_in_memory:
            self.read_in_memory(label_column_name=self.label_names)

    def prepare_data(self, overwrite: bool = False, file_column_name: str = None, make_sdf: bool = True):

        if not os.path.exists(self.file_path_xyz) or overwrite:
            mat = scipy.io.loadmat(os.path.join(self.data_directory, self.download_info["download_file_name"]))
            graph_len = [int(np.around(np.sum(x > 0))) for x in mat["Z"]]
            proton = [x[:i] for i, x in zip(graph_len, mat["Z"])]
            atoms = [[inverse_global_proton_dict[i] for i in x] for x in proton]
            pos = [x[:i, :]*0.529177210903 for i, x in zip(graph_len, mat["R"])]
            atoms_pos = [[x, y] for x, y in zip(atoms, pos)]
            np.save(os.path.join(self.data_directory, "qm7_splits.npy"), mat["P"])
            self.info("Writing XYZ file from coulomb matrix information.")
            write_list_to_xyz_file(self.file_path_xyz, atoms_pos)
        else:
            self.info("Found XYZ file for qm7b already created.")

        if not os.path.exists(self.file_path) or overwrite:
            mat = scipy.io.loadmat(os.path.join(self.data_directory, self.download_info["download_file_name"]))
            labels = mat["T"][0]
            targets = pd.DataFrame(labels, columns=self.label_names)
            self.info("Writing CSV file of graph labels.")
            targets.to_csv(self.file_path, index=False)
        else:
            self.info("Found CSV file of graph labels.")

        return super(QM7Dataset, self).prepare_data(
            overwrite=overwrite, file_column_name=file_column_name, make_sdf=make_sdf)

    def _get_cross_validation_splits(self):
        return np.load(os.path.join(self.data_directory, "qm7_splits.npy"))

    def read_in_memory(self, **kwargs):
        super(QM7Dataset, self).read_in_memory( **kwargs)
        splits = self._get_cross_validation_splits()
        property_split = []
        for i in range(len(self)):
            is_in_split = []
            for j, split in enumerate(splits):
                if i in split:
                    is_in_split.append(j)
            property_split.append(np.array(is_in_split, dtype="int"))
        self.assign_property("kfold", property_split)

        # Mean molecular weight mmw
        mass_dict = {'H': 1.0079, 'C': 12.0107, 'N': 14.0067, 'O': 15.9994, 'F': 18.9984, 'S': 32.065, "C3": 12.0107}

        def mmw(atoms):
            mass = [mass_dict[x[:1]] for x in atoms]
            return np.array([np.mean(mass), len(mass)])

        # TODO: Do this in graph_attributes mol interface.
        self.assign_property("graph_attributes",
                             [mmw(x) if x is not None else None for x in self.obtain_property("node_symbol")])



import os
# import pickle
import numpy as np
import json
import pandas as pd
# from typing import Union
from kgcnn.data.qm import QMDataset
from kgcnn.data.download import DownloadDataset
from kgcnn.mol.io import write_list_to_xyz_file


class QM9Dataset(QMDataset, DownloadDataset):
    r"""Store and process QM9 dataset from `Quantum Machine <http://quantum-machine.org/datasets/>`__ . dataset.

    Dataset of 134k stable small organic molecules made up of C,H,O,N,F.

    From `Quantum Machine <http://quantum-machine.org/datasets/>`__ :
    Computational de novo design of new drugs and materials requires rigorous and unbiased exploration of chemical
    compound space. However, large uncharted territories persist due to its size scaling combinatorially with
    molecular size. We report computed geometric, energetic, electronic, and thermodynamic properties for 134k stable
    small organic molecules made up of CHONF. These molecules correspond to the subset of all 133,885 species with up
    to nine heavy atoms (CONF) out of the GDB-17 chemical universe of 166 billion organic molecules. We report
    geometries minimal in energy, corresponding harmonic frequencies, dipole moments, polarizabilities, along
    with energies, enthalpies, and free energies of atomization. All properties were calculated at
    the B3LYP/6-31G(2df,p) level of quantum chemistry. Furthermore, for the predominant stoichiometry, C7H10O2,
    there are 6,095 constitutional isomers among the 134k molecules. We report energies, enthalpies, and free
    energies of atomization at the more accurate G4MP2 level of theory for all of them. As such, this data set provides
    quantum chemical properties for a relevant, consistent, and comprehensive chemical space of small organic molecules.
    This database may serve the benchmarking of existing methods, development of new methods, such as hybrid quantum
    mechanics/machine learning, and systematic identification of structure-property relationships.

    Labels include geometric, energetic, electronic, and thermodynamic properties. Typically, a random 10% validation
    and 10% test set are used. In literature, test errors are given as MAE and for energies are in [eV].

    Molecules that have a different smiles code after convergence can be removed with :obj:`remove_uncharacterized` .
    Also labels with removed atomization energy are generated.

    .. code-block:: python

        from kgcnn.data.qm import QM9Dataset
        dataset = QM9Dataset(reload=True)
        print(dataset[0])

    References:

        (1) L. Ruddigkeit, R. van Deursen, L. C. Blum, J.-L. Reymond, Enumeration of 166 billion organic small
            molecules in the chemical universe database GDB-17, J. Chem. Inf. Model. 52, 2864–2875, 2012.
        (2) R. Ramakrishnan, P. O. Dral, M. Rupp, O. A. von Lilienfeld, Quantum chemistry structures and properties
            of 134 kilo molecules, Scientific Data 1, 140022, 2014.

    """

    download_info = {
        "dataset_name": "QM9",
        "data_directory_name": "qm9",
        # https://ndownloader.figshare.com/files/3195398
        "download_url": "https://ndownloader.figshare.com/files/3195389",
        "download_file_name": 'dsgdb9nsd.xyz.tar.bz2',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": 'dsgdb9nsd.xyz',
    }

    def __init__(self, reload: bool = False, verbose: int = 10):
        """Initialize QM9 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        QMDataset.__init__(self, verbose=verbose, dataset_name="QM9")
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.label_names = [
            'A', 'B', 'C', 'mu', 'alpha', 'homo', 'lumo', 'gap', 'r2', 'zpve', 'U0', 'U', 'H', 'G', 'Cv',
            'U0_atom', 'U_atom', 'H_atom', 'G_atom', 'Cv_atom']
        self.label_units = [
            "GHz", "GHz", "GHz", "D", r"a_0^3", "eV", "eV", "eV", r"a_0^2", "eV", "eV", "eV", "eV", "eV", r"cal/mol K",
            "eV", "eV", "eV", "eV", r"cal/mol K"]
        self.label_unit_conversion = np.array(
            [[1.0, 1.0, 1.0, 1.0, 1.0, 27.2114, 27.2114, 27.2114, 1.0, 27.2114, 27.2114, 27.2114, 27.2114, 27.2114, 1.0,
              27.2114, 27.2114, 27.2114, 27.2114, 1.0]]
        )  # Pick always same units for training
        self.dataset_name = "QM9"
        self.require_prepare_data = True
        self.fits_in_memory = True
        self.verbose = verbose
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name)
        self.file_name = "qm9.csv"
        self.__removed_uncharacterized = False
        self.__atom_ref = {
            # units are in hartree of original dataset.
            "U0": {"H": -0.500273, "C": -37.846772, "N": -54.583861, "O": -75.064579, "F": -99.718730},
            "U": {"H": -0.498857, "C": -37.845355, "N": -54.582445, "O": -75.063163, "F": -99.717314},
            "H": {"H": -0.497912, "C": -37.844411, "N": -54.581501, "O": -75.062219, "F": -99.716370},
            "G": {"H": -0.510927, "C": -37.861317, "N": -54.598897, "O": -75.079532, "F": -99.733544},
            "CV": {"H": 2.981, "C": 2.981, "N": 2.981, "O": 2.981, "F": 2.981},
        }

        # We try to download also additional files here.
        self.download_database(path=self.data_directory, filename="readme.txt", logger=None,
                               download_url="https://figshare.com/ndownloader/files/3195392", overwrite=reload)
        self.download_database(path=self.data_directory, filename="uncharacterized.txt", logger=None,
                               download_url="https://figshare.com/ndownloader/files/3195404", overwrite=reload)
        self.download_database(path=self.data_directory, filename="atomref.txt", logger=None,
                               download_url="https://figshare.com/ndownloader/files/3195395", overwrite=reload)
        self.download_database(path=self.data_directory, filename="validation.txt", logger=None,
                               download_url="https://figshare.com/ndownloader/files/3195401", overwrite=reload)

        if self.require_prepare_data:
            self.prepare_data(overwrite=reload)

        if self.fits_in_memory:
            self.read_in_memory(
                label_column_name=["%s [%s]" % (a, b) for a, b in zip(self.label_names, self.label_units)])

    def prepare_data(self, overwrite: bool = False, file_column_name: str = None, make_sdf: bool = True):
        """Process data by loading all single xyz-files and store all pickled information to file.
        The single files are deleted afterwards, requires to re-extract the tar-file for overwrite.

        Args:
            overwrite (bool): Whether to redo the processing, requires un-zip of the data again. Defaults to False.
            file_column_name (str): Not used.
            make_sdf (bool): Whether to make SDF file.
        """
        path = self.data_directory
        dataset_size = 133885

        if os.path.exists(self.file_path) and os.path.exists(self.file_path_xyz) and not overwrite:
            self.info("Single XYZ file and CSV table with labels already created.")
            return super(QM9Dataset, self).prepare_data(
                overwrite=overwrite, file_column_name=file_column_name, make_sdf=make_sdf)

        # Reading files.
        if not os.path.exists(os.path.join(path, 'dsgdb9nsd.xyz')):
            self.error("Can not find extracted dsgdb9nsd.xyz directory. Run reload QM9 dataset again.")
            return self

        # Read individual files
        qm9 = []
        self.info("Reading dsgdb9nsd files ...")
        for i in range(1, dataset_size + 1):
            mol = []
            file = "dsgdb9nsd_" + "{:06d}".format(i) + ".xyz"
            open_file = open(os.path.join(path, "dsgdb9nsd.xyz", file), "r")
            lines = open_file.readlines()
            mol.append(int(lines[0]))
            labels = lines[1].strip().split(' ')[1].split('\t')
            if int(labels[0]) != i:
                self.warning("Index for QM9 not matching xyz-file.")
            labels = [lines[1].strip().split(' ')[0].strip()] + [int(labels[0])] + [float(x) for x in labels[1:]]
            mol.append(labels)
            cords = []
            for j in range(int(lines[0])):
                atom_info = lines[2 + j].strip().split('\t')
                cords.append([atom_info[0]] + [float(x.replace('*^', 'e')) for x in atom_info[1:]])
            mol.append(cords)
            freqs = lines[int(lines[0]) + 2].strip().split('\t')
            freqs = [float(x) for x in freqs]
            mol.append(freqs)
            smiles = lines[int(lines[0]) + 3].strip().split('\t')
            mol.append(smiles)
            inchis = lines[int(lines[0]) + 4].strip().split('\t')
            mol.append(inchis)
            open_file.close()
            qm9.append(mol)

        # Save pickled data.
        self.info("Saving qm9.json ...")
        with open(os.path.join(path, "qm9.json"), 'w') as f:
            json.dump(qm9, f)

        # Remove file after reading
        self.info("Cleaning up extracted files...")
        for i in range(1, dataset_size + 1):
            file = "dsgdb9nsd_" + "{:06d}".format(i) + ".xyz"
            file = os.path.join(path, "dsgdb9nsd.xyz", file)
            os.remove(file)

        # Creating XYZ file.
        self.info("Writing single xyz-file ...")
        pos = [[y[1:4] for y in x[2]] for x in qm9]
        atoms = [[y[0] for y in x[2]] for x in qm9]
        atoms_pos = [[x, y] for x, y in zip(atoms, pos)]
        write_list_to_xyz_file(os.path.join(path, "qm9.xyz"), atoms_pos)

        # Creating Table file with labels.
        labels = np.array([x[1][1:] if len(x[1]) == 17 else x[1] for x in qm9])  # Remove 'gdb' tag if not done already.
        atom_energy = [[sum([self.__atom_ref[t][a] for a in x]) for t in ["U0", "U", "H", "G", "CV"]] for x in atoms]
        targets_atom = labels[:, 11:] - np.array(atom_energy)
        targets = np.concatenate([labels[:, 1:], targets_atom], axis=-1)*self.label_unit_conversion
        df = pd.DataFrame(targets, columns=["%s [%s]" % (a, b) for a, b in zip(self.label_names, self.label_units)])
        df.insert(targets.shape[1], "ID", labels[:, 0].astype(dtype="int"))  # add id at the end for reference.
        self.info("Writing CSV file of graph labels.")
        df.to_csv(self.file_path, index=False)

        return super(QM9Dataset, self).prepare_data(
            overwrite=overwrite, file_column_name=file_column_name, make_sdf=make_sdf)

    def read_in_memory(self, **kwargs):
        super(QM9Dataset, self).read_in_memory(**kwargs)

        # Mean molecular weight mmw
        mass_dict = {'H': 1.0079, 'C': 12.0107, 'N': 14.0067, 'O': 15.9994, 'F': 18.9984, 'S': 32.065, "C3": 12.0107}

        def mmw(atoms):
            mass = [mass_dict[x[:1]] for x in atoms]
            return np.array([np.mean(mass), len(mass)])

        # TODO: Do this in graph_attributes mol interface.
        self.assign_property("graph_attributes", [
            mmw(x) if x is not None else None for x in self.obtain_property("node_symbol")])
        return self

    def remove_uncharacterized(self):
        """Remove 3054 uncharacterized molecules that failed structure test from this dataset."""
        if self.__removed_uncharacterized:
            self.error("Uncharacterized molecules have already been removed. Continue.")
            return
        with open(os.path.join(self.data_directory, "uncharacterized.txt"), "r") as f:
            data = f.readlines()[9:-1]
        data = [x.strip().split(" ") for x in data]
        data = [[y for y in x if y != ""] for x in data]
        indices = np.array([x[0] for x in data], dtype="int") - 1
        indices_backward = np.flip(np.sort(indices))
        for i in indices_backward:
            self.pop(int(i))
        self.info("Removed %s uncharacterized molecules." % len(indices_backward))
        self.__removed_uncharacterized = True
        return indices_backward



import numpy as np
from kgcnn.data.datasets.MoleculeNetDataset2018 import MoleculeNetDataset2018


class Tox21MolNetDataset(MoleculeNetDataset2018):
    r"""Store and process 'TOX21' dataset from `MoleculeNet <https://moleculenet.org/>`__ database.

    Compare reference:
    `DeepChem <https://deepchem.readthedocs.io/en/latest/api_reference/moleculenet.html#freesolv-dataset>`__ reading:
    The “Toxicology in the 21st Century” (Tox21) initiative created a public database measuring toxicity of compounds,
    which has been used in the 2014 Tox21 Data Challenge. This dataset contains qualitative toxicity measurements for
    8k compounds on 12 different targets, including nuclear receptors and stress response pathways.

    Random splitting is recommended for this dataset.

    The raw data csv file contains columns below:

        - “smiles”: SMILES representation of the molecular structure
        - “NR-XXX”: Nuclear receptor signaling bioassays results
        - “SR-XXX”: Stress response bioassays results

    please refer to https://tripod.nih.gov/tox21/challenge/data.jsp for details.

    References:

        (1) Tox21 Challenge. https://tripod.nih.gov/tox21/challenge/

    """

    def __init__(self, reload=False, verbose: int = 10, remove_nan: bool = False):
        """Initialize Tox21 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        self._remove_nan_label = remove_nan
        super(Tox21MolNetDataset, self).__init__("Tox21", reload=reload, verbose=verbose)

    def read_in_memory(self, **kwargs):
        super(Tox21MolNetDataset, self).read_in_memory(**kwargs)

        # One could set unknown classes to zero, but better go with NaN compatible metrics and loss.
        graph_labels = self.obtain_property("graph_labels")
        graph_labels = [np.array(x, dtype="float") for x in graph_labels]
        if self._remove_nan_label:
            graph_labels = [np.nan_to_num(x) for x in graph_labels]
        self.assign_property("graph_labels", graph_labels)



from kgcnn.data.datasets.MatBenchDataset2020 import MatBenchDataset2020


class MatProjectGapDataset(MatBenchDataset2020):
    r"""Store and process :obj:`MatProjectGapDataset` from `MatBench <https://matbench.materialsproject.org/>`__
    database.  Name within Matbench: 'matbench_mp_gap'.

    Matbench test dataset for predicting DFT PBE band gap from structure.
    Adapted from Materials Project database. Removed entries having a formation energy (or energy above the convex hull)
    more than 150meV and those containing noble gases. Retrieved April 2, 2019.
    For benchmarking w/ nested cross validation, the order of the dataset must be identical to the retrieved data;
    refer to the Automatminer/Matbench publication for more details.

        * Number of samples: 106113
        * Task type: regression
        * Input type: structure

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize 'matbench_mp_e_form' dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        super(MatProjectGapDataset, self).__init__("matbench_mp_gap", reload=reload, verbose=verbose)
        self.label_names = "gap"
        self.label_units = "eV"



import tensorflow as tf
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.ops.axis import get_positive_axis
from kgcnn.ops.segment import segment_ops_by_name

ks = tf.keras

global_normalization_args = {
    "BatchNormalization": [
        "axis", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "gamma_constraint", "momentum", "moving_mean_initializer",
        "moving_variance_initializer"
    ],
    "GraphBatchNormalization": [
        "axis", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "gamma_constraint", "momentum", "moving_mean_initializer",
        "moving_variance_initializer"
    ],
    "LayerNormalization": [
        "axis", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "gamma_constraint"
    ],
    "GraphLayerNormalization": [
        "axis", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "gamma_constraint"
    ],
    "GraphNormalization": [
        "mean_shift", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "alpha_initializer",
        "beta_regularizer", "gamma_regularizer", "beta_constraint", "alpha_constraint", "gamma_constraint",
        "alpha_regularizer"
    ],
    "GraphInstanceNormalization": [
        "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "alpha_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "alpha_constraint", "gamma_constraint", "alpha_regularizer"
    ],
    "GroupNormalization": [
        "groups", "axis", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "gamma_constraint"]
}


@ks.utils.register_keras_serializable(package='kgcnn', name='GraphLayerNormalization')
class GraphLayerNormalization(GraphBaseLayer):
    r"""Graph Layer normalization for (ragged) graph tensor objects.

    Uses `ks.layers.LayerNormalization` on all node or edge features in a batch.
    Following convention suggested by `GraphNorm: A Principled Approach (...) <https://arxiv.org/abs/2009.03294>`__ .
    To this end, the (positive) :obj:`axis` parameter must be strictly > 0 and ideally > 1,
    since first two dimensions are flattened for normalization.

    The definition of normalization terms for graph neural networks can be categorized as follows. Here we copy the
    definition and description of <https://arxiv.org/abs/2009.03294>`_ . Note that for keras the batch dimension is
    the first dimension.

    .. math::

        \text{Norm}(\hat{h}_{i,j,g}) = \gamma \cdot \frac{\hat{h}_{i,j,g} - \mu}{\sigma} + \beta,


    Consider a batch of graphs :math:`{G_{1}, \dots , G_{b}}` where :math:`b` is the batch size.
    Let :math:`n_{g}` be the number of nodes in graph :math:`G_{g}` .
    We generally denote :math:`\hat{h}_{i,j,g}` as the inputs to the normalization module, e.g.,
    the :math:`j` -th feature value of node :math:`v_i` of graph :math:`G_{g}` ,
    :math:`i = 1, \dots , n_{g}` , :math:`j = 1, \dots , d` , :math:`g = 1, \dots , b` .

    To adapt Layer-Norm to GNNs, we view each node as a basic component, resembling words in a sentence, and apply
    normalization to all feature values across different dimensions of each node,
    i.e. , over dimension :math:`j` of :math:`\hat{h}_{i,j,g}` .

    """

    def __init__(self,
                 axis=-1,
                 epsilon=1e-3, center=True, scale=True,
                 beta_initializer='zeros', gamma_initializer='ones',
                 beta_regularizer=None, gamma_regularizer=None, beta_constraint=None,
                 gamma_constraint=None,
                 **kwargs):
        r"""Initialize layer :obj:`GraphLayerNormalization`.

        Args:
            axis: Integer or List/Tuple. The axis or axes to normalize across.
                Typically this is the features axis/axes. The left-out axes are
                typically the batch axis/axes. This argument defaults to `-1`, the last dimension in the input.
            epsilon: Small float added to variance to avoid dividing by zero. Defaults to 1e-3.
            center: If True, add offset of `beta` to normalized tensor. If False,
                `beta` is ignored. Defaults to True.
            scale: If True, multiply by `gamma`. If False, `gamma` is not used.
                Defaults to True. When the next layer is linear (also e.g. `nn.relu`),
                this can be disabled since the scaling will be done by the next layer.
            beta_initializer: Initializer for the beta weight. Defaults to zeros.
            gamma_initializer: Initializer for the gamma weight. Defaults to ones.
            beta_regularizer: Optional regularizer for the beta weight. None by default.
            gamma_regularizer: Optional regularizer for the gamma weight. None by default.
            beta_constraint: Optional constraint for the beta weight. None by default.
            gamma_constraint: Optional constraint for the gamma weight. None by default.
        """
        super(GraphLayerNormalization, self).__init__(**kwargs)
        # The axis 0,1 are merged for ragged embedding input.
        if isinstance(axis, (list, tuple)):
            self.axis = list(axis[:])
            if any([x == 0 for x in self.axis]):
                raise ValueError("Positive axis for graph normalization must be > 0 or negative.")
            axis_values = [x - 1 if x > 0 else x for x in self.axis]
        elif isinstance(axis, int):
            self.axis = axis
            if self.axis == 0:
                raise ValueError("Positive axis for graph normalization must be > 0 or negative.")
            axis_values = axis - 1 if axis > 0 else axis
        else:
            raise TypeError('Expected an int or a list/tuple of ints for the '
                            'argument \'axis\', but received: %r' % axis)

        self._layer_norm = ks.layers.LayerNormalization(axis=axis_values, epsilon=epsilon,
                                                        center=center, scale=scale,
                                                        beta_initializer=beta_initializer,
                                                        gamma_initializer=gamma_initializer,
                                                        beta_regularizer=beta_regularizer,
                                                        gamma_regularizer=gamma_regularizer,
                                                        beta_constraint=beta_constraint,
                                                        gamma_constraint=gamma_constraint, dtype=self.dtype)
        # We can add config from keras layer except the axis parameter.
        self._add_layer_config_to_self = {"_layer_norm": ["epsilon", "center", "scale", "beta_initializer",
                                                          "gamma_initializer", "beta_regularizer", "gamma_regularizer",
                                                          "beta_constraint", "gamma_constraint"]}

    def build(self, input_shape):
        """Build layer."""
        super(GraphLayerNormalization, self).build(input_shape)
        n_dims = len(input_shape)
        if isinstance(self.axis, int):
            axis = get_positive_axis(self.axis, n_dims)
            if axis < 1:
                raise ValueError("The (positive) axis must be > 0.")
            axis_values = axis - 1
        elif isinstance(self.axis, list):
            axis = [get_positive_axis(x, n_dims) for x in self.axis]
            if any([x < 1 for x in axis]):
                raise ValueError("All (positive) axis must be > 0.")
            axis_values = [x - 1 for x in axis]
        else:
            raise TypeError("Expected an int or a list of ints for the axis %s" % self.axis)
        # Give positive axis to self, after built the axis is always positive
        self.axis = axis
        # Remove batch dimension as we will call directly on value tensor in call.
        self._layer_norm.axis = axis_values

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor, tf.Tensor): Embeddings of shape (batch, [M], F, ...)

        Returns:
            tf.RaggedTensor: Normalized ragged tensor of identical shape (batch, [M], F, ...)
        """
        inputs = self.assert_ragged_input_rank(inputs, ragged_rank=1)  # Must have ragged_rank = 1.
        return self.map_values(self._layer_norm, inputs, **kwargs)

    def get_config(self):
        """Get layer configuration."""
        config = super(GraphLayerNormalization, self).get_config()
        config.update({"axis": self.axis})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='GraphBatchNormalization')
class GraphBatchNormalization(GraphBaseLayer):
    r"""Graph batch normalization for (ragged) graph tensor objects.

    Uses `ks.layers.BatchNormalization` on all node or edge features in a batch.
    Following convention suggested by `GraphNorm: A Principled Approach (...) <https://arxiv.org/abs/2009.03294>`__ .
    To this end, the (positive) :obj:`axis` parameter must be strictly > 0 and ideally > 1,
    since first two dimensions are flattened for normalization.

    The definition of normalization terms for graph neural networks can be categorized as follows. Here we copy the
    definition and description of `<https://arxiv.org/abs/2009.03294>`_ . Note that for keras the batch dimension is
    the first dimension.

    .. math::

        \text{Norm}(\hat{h}_{i,j,g}) = \gamma \cdot \frac{\hat{h}_{i,j,g} - \mu}{\sigma} + \beta,


    Consider a batch of graphs :math:`{G_{1}, \dots , G_{b}}` where :math:`b` is the batch size.
    Let :math:`n_{g}` be the number of nodes in graph :math:`G_{g}` .
    We generally denote :math:`\hat{h}_{i,j,g}` as the inputs to the normalization module, e.g.,
    the :math:`j` -th feature value of node :math:`v_i` of graph :math:`G_{g}` ,
    :math:`i = 1, \dots , n_{g}` , :math:`j = 1, \dots , d` , :math:`g = 1, \dots , b` .

    For BatchNorm, normalization and the computation of :math:`mu`
    and :math:`\sigma` are applied to all values in the same feature dimension
    across the nodes of all graphs in the batch as in
    `Xu et al. (2019) <https://openreview.net/forum?id=ryGs6iA5Km>`__ , i.e., over dimensions :math:`g`, :math:`i`
    of :math:`\hat{h}_{i,j,g}` .

    """

    def __init__(self,
                 axis=-1,
                 momentum=0.99, epsilon=0.001, center=True, scale=True,
                 beta_initializer='zeros', gamma_initializer='ones',
                 moving_mean_initializer='zeros',
                 moving_variance_initializer='ones', beta_regularizer=None,
                 gamma_regularizer=None, beta_constraint=None, gamma_constraint=None,
                 **kwargs):
        r"""Initialize layer :obj:`GraphBatchNormalization`.

        Args:
            axis: Integer, the axis that should be normalized (typically the features
                axis). For instance, after a `Conv2D` layer with
                `data_format="channels_first"`, set `axis=1` in `BatchNormalization`.
            momentum: Momentum for the moving average.
            epsilon: Small float added to variance to avoid dividing by zero.
            center: If True, add offset of `beta` to normalized tensor. If False,
                `beta` is ignored.
            scale: If True, multiply by `gamma`. If False, `gamma` is not used. When
                the next layer is linear (also e.g. `nn.relu`), this can be disabled
                since the scaling will be done by the next layer.
            beta_initializer: Initializer for the beta weight.
            gamma_initializer: Initializer for the gamma weight.
            moving_mean_initializer: Initializer for the moving mean.
            moving_variance_initializer: Initializer for the moving variance.
            beta_regularizer: Optional regularizer for the beta weight.
            gamma_regularizer: Optional regularizer for the gamma weight.
            beta_constraint: Optional constraint for the beta weight.
            gamma_constraint: Optional constraint for the gamma weight.

        """
        super(GraphBatchNormalization, self).__init__(**kwargs)
        # The axis 0,1 are merged for ragged embedding input.
        if isinstance(axis, (list, tuple)):
            self.axis = list(axis[:])
            if any([x == 0 for x in self.axis]):
                raise ValueError("Positive axis for graph normalization must be > 0 or negative.")
            axis_values = [x - 1 if x > 0 else x for x in self.axis]
        elif isinstance(axis, int):
            self.axis = axis
            if self.axis == 0:
                raise ValueError("Positive axis for graph normalization must be > 0 or negative.")
            axis_values = axis - 1 if axis > 0 else axis
        else:
            raise TypeError('Expected an int or a list/tuple of ints for the '
                            'argument \'axis\', but received: %r' % axis)

        self._layer_norm = ks.layers.BatchNormalization(axis=axis_values, momentum=momentum, epsilon=epsilon,
                                                        center=center, scale=scale,
                                                        beta_initializer=beta_initializer,
                                                        gamma_initializer=gamma_initializer,
                                                        moving_mean_initializer=moving_mean_initializer,
                                                        moving_variance_initializer=moving_variance_initializer,
                                                        beta_regularizer=beta_regularizer,
                                                        gamma_regularizer=gamma_regularizer,
                                                        beta_constraint=beta_constraint,
                                                        gamma_constraint=gamma_constraint)
        # We can add config from keras layer except the axis parameter.
        self._add_layer_config_to_self = {
            "_layer_norm": ["momentum", "epsilon", "scale", "center", "beta_initializer", "gamma_initializer",
                            "moving_mean_initializer", "moving_variance_initializer"
                                                       "beta_regularizer", "gamma_regularizer", "beta_constraint",
                            "gamma_constraint"]}

    def build(self, input_shape):
        """Build layer."""
        super(GraphBatchNormalization, self).build(input_shape)
        n_dims = len(input_shape)
        if isinstance(self.axis, int):
            axis = get_positive_axis(self.axis, n_dims)
            if axis < 1:
                raise ValueError("The (positive) axis must be > 0.")
            axis_values = axis - 1
        elif isinstance(self.axis, list):
            axis = [get_positive_axis(x, n_dims) for x in self.axis]
            if any([x < 1 for x in axis]):
                raise ValueError("All (positive) axis must be > 0.")
            axis_values = [x - 1 for x in axis]
        else:
            raise TypeError("Expected an int or a list of ints for the axis %s" % self.axis)
        # Give positive axis to self, after built the axis is always positive
        self.axis = axis
        # Remove batch dimension as we will call directly on value tensor in call.
        self._layer_norm.axis = axis_values

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor, tf.Tensor): Embeddings of shape (batch, [M], F, ...)

        Returns:
            tf.RaggedTensor: Normalized ragged tensor of identical shape (batch, [M], F, ...)
        """
        inputs = self.assert_ragged_input_rank(inputs, ragged_rank=1)  # Must have ragged_rank = 1.
        return self.map_values(self._layer_norm, inputs, **kwargs)

    def get_config(self):
        """Get layer configuration."""
        config = super(GraphBatchNormalization, self).get_config()
        config.update({"axis": self.axis})
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='GraphNormalization')
class GraphNormalization(GraphBaseLayer):
    r"""Graph normalization for (ragged) graph tensor objects.

    Following convention suggested by `GraphNorm: A Principled Approach (...) <https://arxiv.org/abs/2009.03294>`__ .

    The definition of normalization terms for graph neural networks can be categorized as follows. Here we copy the
    definition and description of `<https://arxiv.org/abs/2009.03294>`_ . Note that for keras the batch dimension is
    the first dimension.

    .. math::

        \text{Norm}(\hat{h}_{i,j,g}) = \gamma \cdot \frac{\hat{h}_{i,j,g} - \mu}{\sigma} + \beta,


    Consider a batch of graphs :math:`{G_{1}, \dots , G_{b}}` where :math:`b` is the batch size.
    Let :math:`n_{g}` be the number of nodes in graph :math:`G_{g}` .
    We generally denote :math:`\hat{h}_{i,j,g}` as the inputs to the normalization module, e.g.,
    the :math:`j` -th feature value of node :math:`v_i` of graph :math:`G_{g}` ,
    :math:`i = 1, \dots , n_{g}` , :math:`j = 1, \dots , d` , :math:`g = 1, \dots , b` .

    For InstanceNorm, we regard each graph as an instance. The normalization is
    then applied to the feature values across all nodes for each
    individual graph, i.e., over dimension :math:`i` of :math:`\hat{h}_{i,j,g}` .

    Additionally, the following proposed additions for GraphNorm are added when compared to InstanceNorm.

    .. math::

        \text{GraphNorm}(\hat{h}_{i,j}) = \gamma_j \cdot \frac{\hat{h}_{i,j} - \alpha_j \mu_j }{\hat{\sigma}_j}+\beta_j

    where :math:`\mu_j = \frac{\sum^n_{i=1} hat{h}_{i,j}}{n}` ,
    :math:`\hat{\sigma}^2_j = \frac{\sum^n_{i=1} (hat{h}_{i,j} - \alpha_j \mu_j)^2}{n}`  ,
    and :math:`\gamma_j` , :math:`beta_j` are the affine parameters as in other normalization methods.

    .. code-block:: python

        import tensorflow as tf
        from kgcnn.layers.norm import GraphNormalization
        layer = GraphNormalization()
        test = tf.ragged.constant([[[0.0, 0.0],[1.0, -1.0]],[[1.0, 1.0],[0.0, 0.0],[-2.0, -2.0]]], ragged_rank=1)
        print(layer(test))

    """

    def __init__(self,
                 mean_shift=True, epsilon=1e-3, center=True, scale=True,
                 beta_initializer='zeros', gamma_initializer='ones', alpha_initializer='ones',
                 beta_regularizer=None, gamma_regularizer=None, alpha_regularizer=None,
                 beta_constraint=None, gamma_constraint=None, alpha_constraint=None,
                 **kwargs):
        r"""Initialize layer :obj:`GraphBatchNormalization`.

        Args:
            epsilon: Small float added to variance to avoid dividing by zero. Defaults to 1e-3.
            center: If True, add offset of `beta` to normalized tensor. If False,
                `beta` is ignored. Defaults to True.
            scale: If True, multiply by `gamma`. If False, `gamma` is not used.
                Defaults to True. When the next layer is linear (also e.g. `nn.relu`),
                this can be disabled since the scaling will be done by the next layer.
            mean_shift (bool): Whether to apply alpha. Default is True.
            beta_initializer: Initializer for the beta weight. Defaults to 'zeros'.
            gamma_initializer: Initializer for the gamma weight. Defaults to 'ones'.
            alpha_initializer: Initializer for the alpha weight. Defaults to 'ones'.
            beta_regularizer: Optional regularizer for the beta weight. None by default.
            gamma_regularizer: Optional regularizer for the gamma weight. None by default.
            alpha_regularizer: Optional regularizer for the alpha weight. None by default.
            beta_constraint: Optional constraint for the beta weight. None by default.
            gamma_constraint: Optional constraint for the gamma weight. None by default.
            alpha_constraint: Optional constraint for the alpha weight. None by default.
        """
        super(GraphNormalization, self).__init__(**kwargs)
        self.epsilon = epsilon
        self._eps = tf.constant(epsilon, dtype=self.dtype)
        self.center = center
        self.mean_shift = mean_shift
        self.scale = scale
        self.beta_initializer = ks.initializers.get(beta_initializer)
        self.gamma_initializer = ks.initializers.get(gamma_initializer)
        self.alpha_initializer = ks.initializers.get(alpha_initializer)
        self.beta_regularizer = ks.regularizers.get(beta_regularizer)
        self.gamma_regularizer = ks.regularizers.get(gamma_regularizer)
        self.alpha_regularizer = ks.regularizers.get(alpha_regularizer)
        self.beta_constraint = ks.constraints.get(beta_constraint)
        self.gamma_constraint = ks.constraints.get(gamma_constraint)
        self.alpha_constraint = ks.constraints.get(alpha_constraint)
        # Weights
        self.alpha = None
        self.gamma = None
        self.beta = None

    def build(self, input_shape):
        """Build layer."""
        super(GraphNormalization, self).build(input_shape)
        param_shape = [x if x is not None else 1 for x in input_shape[2:]]
        if self.scale:
            self.gamma = self.add_weight(
                name="gamma",
                shape=param_shape,
                initializer=self.gamma_initializer,
                regularizer=self.gamma_regularizer,
                constraint=self.gamma_constraint,
                trainable=True,
                experimental_autocast=False,
            )

        if self.center:
            self.beta = self.add_weight(
                name="beta",
                shape=param_shape,
                initializer=self.beta_initializer,
                regularizer=self.beta_regularizer,
                constraint=self.beta_constraint,
                trainable=True,
                experimental_autocast=False,
            )

        if self.mean_shift:
            self.alpha = self.add_weight(
                name="alpha",
                shape=param_shape,
                initializer=self.alpha_initializer,
                regularizer=self.alpha_regularizer,
                constraint=self.alpha_constraint,
                trainable=True,
                experimental_autocast=False,
            )

        self.built = True

    def _ragged_mean_std(self, inputs):
        # Here a segment operation for ragged_rank=1 tensors is used.
        # Alternative is to simply use tf.reduce_mean which should also work for latest tf-version.
        # Then tf.nn.moments could be used or similar tf implementation for variance and mean.
        values = inputs.values
        if values.dtype in ("float16", "bfloat16") and self.dtype == "float32":
            values = tf.cast(values, "float32")
        mean = segment_ops_by_name("mean", values, inputs.value_rowids())
        if self.mean_shift:
            mean = mean * tf.expand_dims(self.alpha, axis=0)
        mean = tf.repeat(mean, inputs.row_lengths(), axis=0)
        diff = values - mean
        # Not sure whether to stop gradients for variance if alpha ist used.
        square_diff = tf.square(diff)  # values - tf.stop_gradient(mean)
        variance = segment_ops_by_name("mean", square_diff, inputs.value_rowids())
        std = tf.sqrt(variance + self._eps)
        std = tf.repeat(std, inputs.row_lengths(), axis=0)
        return mean, std, diff / std

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor, tf.Tensor): Node or edge embeddings of shape (batch, [M], F, ...)

        Returns:
            tf.RaggedTensor: Normalized ragged tensor of identical shape (batch, [M], F, ...)
        """
        inputs = self.assert_ragged_input_rank(inputs, ragged_rank=1)  # Must have ragged_rank = 1.
        mean, std, new_values = self._ragged_mean_std(inputs)
        # Recomputing diff.
        if self.scale:
            new_values = new_values * tf.expand_dims(self.gamma, axis=0)
        if self.center:
            new_values = new_values + self.beta
        return inputs.with_values(new_values)

    def get_config(self):
        """Get layer configuration."""
        config = super(GraphNormalization, self).get_config()
        config.update({
            "mean_shift": self.mean_shift,
            "epsilon": self.epsilon,
            "center": self.center,
            "scale": self.scale,
            "beta_initializer": ks.initializers.serialize(self.beta_initializer),
            "gamma_initializer": ks.initializers.serialize(self.gamma_initializer),
            "alpha_initializer": ks.initializers.serialize(self.alpha_initializer),
            "beta_regularizer": ks.regularizers.serialize(self.beta_regularizer),
            "gamma_regularizer": ks.regularizers.serialize(self.gamma_regularizer),
            "alpha_regularizer": ks.regularizers.serialize(self.alpha_regularizer),
            "beta_constraint": ks.constraints.serialize(self.beta_constraint),
            "gamma_constraint": ks.constraints.serialize(self.gamma_constraint),
            "alpha_constraint": ks.constraints.serialize(self.alpha_constraint),
        })
        return config


@ks.utils.register_keras_serializable(package='kgcnn', name='GraphInstanceNormalization')
class GraphInstanceNormalization(GraphNormalization):
    r"""Graph instance normalization for (ragged) graph tensor objects.

    Following convention suggested by `GraphNorm: A Principled Approach (...) <https://arxiv.org/abs/2009.03294>`__ .

    The definition of normalization terms for graph neural networks can be categorized as follows. Here we copy the
    definition and description of `<https://arxiv.org/abs/2009.03294>`_ . Note that for keras the batch dimension is
    the first dimension.

    .. math::

        \text{Norm}(\hat{h}_{i,j,g}) = \gamma \cdot \frac{\hat{h}_{i,j,g} - \mu}{\sigma} + \beta,

    Consider a batch of graphs :math:`{G_{1}, \dots , G_{b}}` where :math:`b` is the batch size.
    Let :math:`n_{g}` be the number of nodes in graph :math:`G_{g}` .
    We generally denote :math:`\hat{h}_{i,j,g}` as the inputs to the normalization module, e.g.,
    the :math:`j` -th feature value of node :math:`v_i` of graph :math:`G_{g}` ,
    :math:`i = 1, \dots , n_{g}` , :math:`j = 1, \dots , d` , :math:`g = 1, \dots , b` .

    For InstanceNorm, we regard each graph as an instance. The normalization is
    then applied to the feature values across all nodes for each
    individual graph, i.e., over dimension :math:`i` of :math:`\hat{h}_{i,j,g}` .

    .. code-block:: python

        import tensorflow as tf
        from kgcnn.layers.norm import GraphInstanceNormalization
        layer = GraphInstanceNormalization()
        test = tf.ragged.constant([[[0.0, 0.0],[1.0, -1.0]],[[1.0, 1.0],[0.0, 0.0],[-2.0, -2.0]]], ragged_rank=1)
        print(layer(test))

    """

    def __init__(self, **kwargs):
        r"""Initialize layer :obj:`GraphBatchNormalization`.

        Args:
            epsilon: Small float added to variance to avoid dividing by zero. Defaults to 1e-3.
            center: If True, add offset of `beta` to normalized tensor. If False,
                `beta` is ignored. Defaults to True.
            scale: If True, multiply by `gamma`. If False, `gamma` is not used.
                Defaults to True. When the next layer is linear (also e.g. `nn.relu`),
                this can be disabled since the scaling will be done by the next layer.
            beta_initializer: Initializer for the beta weight. Defaults to 'zeros'.
            gamma_initializer: Initializer for the gamma weight. Defaults to 'ones'.
            alpha_initializer: Initializer for the alpha weight. Defaults to 'ones'.
            beta_regularizer: Optional regularizer for the beta weight. None by default.
            gamma_regularizer: Optional regularizer for the gamma weight. None by default.
            alpha_regularizer: Optional regularizer for the alpha weight. None by default.
            beta_constraint: Optional constraint for the beta weight. None by default.
            gamma_constraint: Optional constraint for the gamma weight. None by default.
            alpha_constraint: Optional constraint for the alpha weight. None by default.

        """
        super(GraphInstanceNormalization, self).__init__(mean_shift=False, **kwargs)



import numpy as np
import tensorflow as tf
ks = tf.keras


@ks.utils.register_keras_serializable(package='kgcnn', name='ScaledMeanAbsoluteError')
class ScaledMeanAbsoluteError(ks.metrics.MeanAbsoluteError):
    """Metric for a scaled mean absolute error (MAE), which can undo a pre-scaling of the targets. Only intended as
    metric this allows to info the MAE with correct units or absolute values during fit."""

    def __init__(self, scaling_shape=(), name='mean_absolute_error', **kwargs):
        super(ScaledMeanAbsoluteError, self).__init__(name=name, **kwargs)
        self.scale = self.add_weight(shape=scaling_shape,
                                     initializer=ks.initializers.Ones(),
                                     name='kgcnn_scale_mae',
                                     dtype=ks.backend.floatx(),
                                     synchronization=tf.VariableSynchronization.AUTO,
                                     aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
                                     )
        self.scaling_shape = scaling_shape

    def reset_state(self):
        ks.backend.batch_set_value([(v, 0) for v in self.variables if 'kgcnn_scale_mae' not in v.name])

    def reset_states(self):
        ks.backend.batch_set_value([(v, 0) for v in self.variables if 'kgcnn_scale_mae' not in v.name])

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true = self.scale * y_true
        y_pred = self.scale * y_pred
        return super(ScaledMeanAbsoluteError, self).update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self):
        """Returns the serializable config of the metric."""
        mae_conf = super(ScaledMeanAbsoluteError, self).get_config()
        mae_conf.update({"scaling_shape": self.scaling_shape})
        return mae_conf

    def set_scale(self, scale):
        """Set the scale from numpy array. Usually used with broadcasting."""
        ks.backend.set_value(self.scale, scale)

    def merge_state(self, metrics):
        assign_add_ops = []
        for metric in metrics:
            if len(self.weights) != len(metric.weights):
                raise ValueError(f'Metric {metric} is not compatible with {self}')
            for weight, weight_to_add in zip(self.weights, metric.weights):
                if 'kgcnn_scale_mae' not in weight.name:
                    assign_add_ops.append(weight.assign_add(weight_to_add))
        return assign_add_ops


@ks.utils.register_keras_serializable(package='kgcnn', name='RaggedScaledMeanAbsoluteError')
class RaggedScaledMeanAbsoluteError(ks.metrics.MeanAbsoluteError):
    """Metric for a scaled mean absolute error (MAE), which can undo a pre-scaling of the targets. Only intended as
    metric this allows to info the MAE with correct units or absolute values during fit."""

    def __init__(self, scaling_shape=(), name='mean_absolute_error', **kwargs):
        super(RaggedScaledMeanAbsoluteError, self).__init__(name=name, **kwargs)
        self.scale = self.add_weight(shape=scaling_shape,
                                     initializer=ks.initializers.Ones(),
                                     name='kgcnn_scale_mae',
                                     dtype=ks.backend.floatx(),
                                     synchronization=tf.VariableSynchronization.AUTO,
                                     aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
                                     )
        self.scaling_shape = scaling_shape

    def reset_state(self):
        ks.backend.batch_set_value([(v, 0) for v in self.variables if 'kgcnn_scale_mae' not in v.name])

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true, y_pred = [tf.ragged.map_flat_values(lambda x: x * self.scale, y) for y in [y_true, y_pred]]
        # return super(RaggedScaledMeanAbsoluteError, self).update_state(
        #   y_true.flat_values, y_pred.flat_values, sample_weight=sample_weight)
        return super(RaggedScaledMeanAbsoluteError, self).update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self):
        """Returns the serializable config of the metric."""
        mae_conf = super(RaggedScaledMeanAbsoluteError, self).get_config()
        mae_conf.update({"scaling_shape": self.scaling_shape})
        return mae_conf

    def set_scale(self, scale):
        """Set the scale from numpy array. Usually used with broadcasting."""
        ks.backend.set_value(self.scale, scale)

    def merge_state(self, metrics):
        assign_add_ops = []
        for metric in metrics:
            if len(self.weights) != len(metric.weights):
                raise ValueError(f'Metric {metric} is not compatible with {self}')
            for weight, weight_to_add in zip(self.weights, metric.weights):
                if 'kgcnn_scale_mae' not in weight.name:
                    assign_add_ops.append(weight.assign_add(weight_to_add))
        return assign_add_ops


@ks.utils.register_keras_serializable(package='kgcnn', name='ScaledRootMeanSquaredError')
class ScaledRootMeanSquaredError(ks.metrics.RootMeanSquaredError):
    """Metric for a scaled root mean squared error (RMSE), which can undo a pre-scaling of the targets.
    Only intended as metric this allows to info the MAE with correct units or absolute values during fit."""

    def __init__(self, scaling_shape=(), name='root_mean_squared_error', **kwargs):
        super(ScaledRootMeanSquaredError, self).__init__(name=name, **kwargs)
        self.scale = self.add_weight(shape=scaling_shape,
                                     initializer=ks.initializers.Ones(),
                                     name='kgcnn_scale_rmse',
                                     dtype=ks.backend.floatx(),
                                     synchronization=tf.VariableSynchronization.AUTO,
                                     aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
                                     )
        self.scaling_shape = scaling_shape

    def reset_state(self):
        ks.backend.batch_set_value([(v, 0) for v in self.variables if 'kgcnn_scale_rmse' not in v.name])

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true = self.scale * y_true
        y_pred = self.scale * y_pred
        return super(ScaledRootMeanSquaredError, self).update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self):
        """Returns the serializable config of the metric."""
        mae_conf = super(ScaledRootMeanSquaredError, self).get_config()
        mae_conf.update({"scaling_shape": self.scaling_shape})
        return mae_conf

    def set_scale(self, scale):
        """Set the scale from numpy array. Usually used with broadcasting."""
        ks.backend.set_value(self.scale, scale)

    def merge_state(self, metrics):
        assign_add_ops = []
        for metric in metrics:
            if len(self.weights) != len(metric.weights):
                raise ValueError(f'Metric {metric} is not compatible with {self}')
            for weight, weight_to_add in zip(self.weights, metric.weights):
                if 'kgcnn_scale_rmse' not in weight.name:
                    assign_add_ops.append(weight.assign_add(weight_to_add))
        return assign_add_ops


@ks.utils.register_keras_serializable(package='kgcnn', name='BinaryAccuracyNoNaN')
class BinaryAccuracyNoNaN(ks.metrics.BinaryAccuracy):

    def __init__(self, name="binary_accuracy_no_nan", **kwargs):
        super(BinaryAccuracyNoNaN, self).__init__(name=name, **kwargs)

    def update_state(self, y_true, y_pred, sample_weight=None):
        is_not_nan = tf.cast(tf.logical_not(tf.math.is_nan(y_true)), y_pred.dtype)
        if sample_weight is not None:
            sample_weight *= is_not_nan
        else:
            sample_weight = is_not_nan
        return super(BinaryAccuracyNoNaN, self).update_state(y_true, y_pred, sample_weight=sample_weight)


@ks.utils.register_keras_serializable(package='kgcnn', name='AUCNoNaN')
class AUCNoNaN(ks.metrics.AUC):

    def __init__(self, name="AUC_no_nan", **kwargs):
        super(AUCNoNaN, self).__init__(name=name, **kwargs)

    def update_state(self, y_true, y_pred, sample_weight=None):
        is_not_nan = tf.cast(tf.logical_not(tf.math.is_nan(y_true)), y_pred.dtype)
        if sample_weight is not None:
            sample_weight *= is_not_nan
        else:
            sample_weight = is_not_nan
        return super(AUCNoNaN, self).update_state(y_true, y_pred, sample_weight=sample_weight)


@ks.utils.register_keras_serializable(package='kgcnn', name='BalancedBinaryAccuracyNoNaN')
class BalancedBinaryAccuracyNoNaN(ks.metrics.Metric):

    def __init__(self, name="balanced_binary_accuracy_no_nan", thresholds=None, **kwargs):
        super(BalancedBinaryAccuracyNoNaN, self).__init__(name=name, **kwargs)
        self.thresholds = thresholds
        self.confusion_metrics = [
            tf.keras.metrics.TruePositives(thresholds=thresholds, **kwargs),
            tf.keras.metrics.FalsePositives(thresholds=thresholds, **kwargs),
            tf.keras.metrics.TrueNegatives(thresholds=thresholds, **kwargs),
            tf.keras.metrics.FalseNegatives(thresholds=thresholds, **kwargs)
        ]

    def update_state(self, y_true, y_pred, sample_weight=None):
        """Update the state of the metric.

        Args:
            y_true: Ground truth label values. shape = `[batch_size, d0, .. dN-1]`
                or shape = `[batch_size, d0, .. dN-1, 1]` .
            y_pred: The predicted probability values. shape = `[batch_size, d0, .. dN]` .
            sample_weight: Optional sample_weight acts as a coefficient for the metric.

        Returns:
            update ops.
        """

        is_not_nan = tf.cast(tf.logical_not(tf.math.is_nan(y_true)), y_pred.dtype)
        if sample_weight is not None:
            sample_weight *= is_not_nan
        else:
            sample_weight = is_not_nan

        update_ops = [m.update_state(y_true, y_pred, sample_weight=sample_weight) for m in self.confusion_metrics]
        return tf.group(update_ops)

    def result(self):
        true_positives, false_positives, true_negatives, false_negatives = [
            m.result() for m in self.confusion_metrics
        ]
        result = (tf.math.divide_no_nan(true_positives, true_positives + false_negatives) +
                  tf.math.divide_no_nan(true_negatives, true_negatives + false_positives)) / 2
        return result

    def reset_state(self):
        for m in self.confusion_metrics:
            m.reset_state()

    def get_config(self):
        config = super(BalancedBinaryAccuracyNoNaN, self).get_config()
        config.update({
            "thresholds": self.thresholds,
        })
        return config



import unittest
import tempfile
import os
import random
from pprint import pprint

import numpy as np
import pandas as pd
from rdkit import RDLogger
RDLogger.DisableLog('rdApp.*')

from kgcnn.mol.graph_rdkit import MolecularGraphRDKit
from kgcnn.data.moleculenet import MoleculeNetDataset, map_molecule_callbacks
from kgcnn.mol.encoder import OneHotEncoder

from .utils import ASSETS_PATH


class TestMoleculeNetDataset(unittest.TestCase):

    # -- UNITTESTS --

    def test_construction_basically_works(self):
        """
        If it is possible to construct a new instance of MoleculeNetDataset without an error
        """
        csv_path = os.path.join('simple_molecules.csv')

        with tempfile.TemporaryDirectory() as path:
            molecule_net = MoleculeNetDataset(
                data_directory=path,
                file_name=csv_path,
                dataset_name='test'
            )
            self.assertIsInstance(molecule_net, MoleculeNetDataset)

    def test_basically_works(self):
        """
        If one typical processing pipeline can be executed without errors. This includes constructing a
        new instance and then loading the processed molecules from the csv file into memory as GraphList
        instances.
        """
        csv_path = os.path.join(ASSETS_PATH, 'simple_molecules.csv')

        with tempfile.TemporaryDirectory() as path:
            molecule_net = MoleculeNetDataset(
                data_directory=path,
                file_name=csv_path,
                dataset_name='test'
            )

            molecule_net.prepare_data(
                overwrite=False,
                smiles_column_name='smiles'
            )

            molecule_net.read_in_memory(
                label_column_name='label',
                add_hydrogen=False,
            )

            # We know that this csv file has 4 actual data rows
            self.assertEqual(len(molecule_net), 4)

            for molecule in molecule_net:
                self.assertIsInstance(molecule, dict)

    def test_faulty_smiles_do_not_cause_exception(self):
        """
        Even if the source CSV file contains faulty smiles codes, the full molecule net processing should
        run without an exception interrupting the runtime.
        """
        csv_path = os.path.join(ASSETS_PATH, 'faulty_molecules.csv')

        with tempfile.TemporaryDirectory() as path:
            molecule_net = MoleculeNetDataset(
                data_directory=path,
                file_name=csv_path,
                dataset_name='test'
            )

            molecule_net.prepare_data(
                overwrite=False,
                smiles_column_name='smiles'
            )

            molecule_net.read_in_memory(
                label_column_name='label',
                add_hydrogen=False,
            )

            # At this point the test is successful if there was no error, which is what we wanted to
            # show with this test case: Even if the CSV contains faulty SMILES there will be no error
            # which prevents the processing of the dataset
            self.assertEqual(len(molecule_net), 4)

    def test_pandas_from_csv(self):
        """
        Simply test if pandas data frame indexing works as expected.
        """
        csv_path = os.path.join(ASSETS_PATH, 'simple_molecules.csv')
        data = pd.read_csv(csv_path)

        row_index = 0
        # This line is supposed to return a dict for the values of the row of the CSV file with the specified
        # index
        row_dict = dict(data.loc[row_index])
        self.assertEqual(len(row_dict), 4)
        self.assertIn('index', row_dict.keys())
        self.assertIn('label', row_dict.keys())
        self.assertIn('smiles', row_dict.keys())
        self.assertIn('name', row_dict.keys())

    def test_map_molecule_callbacks_basically_works(self):
        """
        Tests if the new function "map_molecule_callbacks" basically works as intended, which is the
        possibility to define custom callbacks based on the CSV data and the molecule object instance
        which add custom properties to the final GraphList object loaded by the dataest.
        """
        csv_path = os.path.join(ASSETS_PATH, 'simple_molecules.csv')

        with tempfile.TemporaryDirectory() as path:

            molecule_net = MoleculeNetDataset(
                data_directory=path,
                file_name=csv_path,
                dataset_name='test'
            )

            callbacks = {
                'name': lambda mg, dd: np.array(dd['name'], dtype='str'),
                'label': lambda mg, dd: np.array(dd['label'], dtype='str')
            }

            # Before we can call the method, the data needs to be prepared because we need to make sure
            # that the molecule file exists before attempting to read from it.
            molecule_net.prepare_data(
                overwrite=False,
                smiles_column_name='smiles'
            )

            # This method will automatically add properties with the string key names of the "callbacks"
            # dict to the underlying GraphList based on the functionality defined by the callback functions.
            mol_values = map_molecule_callbacks(
                mol_list=molecule_net.get_mol_blocks_from_sdf_file(),
                data=molecule_net.read_in_table_file().data_frame,
                callbacks=callbacks,
                mol_interface_class=MolecularGraphRDKit
            )

            for key, value in mol_values.items():
                molecule_net.assign_property(key, value)

            molecule = molecule_net[1]
            self.assertIn('name', molecule)
            self.assertIn('label', molecule)

            # But the other default attributes are not part of it, since we did not specify those in the
            # callbacks dict
            self.assertNotIn('node_symbol', molecule)

    def test_graph_labels_is_one_dimensional_array(self):
        """
        If the shape of the "graph_labels" property of the loaded molecules behaves as expected, which is
        a 0-dimensional np array if only a single target value column is given and a 1-dimensional array
        if multiple columns are given.
        """
        csv_path = os.path.join(ASSETS_PATH, 'simple_molecules.csv')

        # In this first setting we pass a single string as the label_column_name.
        # The corresponding "graph_labels"
        # attribute of the dataset elements should be a 0-dimensional np array containing the values
        # corresponding to the CSV column of that name
        with tempfile.TemporaryDirectory() as path:
            molecule_net = MoleculeNetDataset(
                data_directory=path,
                file_name=csv_path,
                dataset_name='test'
            )
            molecule_net.prepare_data()
            molecule_net.read_in_memory(label_column_name='label')
            molecule_net.set_attributes()

            molecule = molecule_net[0]
            self.assertIsInstance(molecule['graph_labels'], np.ndarray)
            # It is just a single value, aka 0-dimensional array
            self.assertEqual(molecule['graph_labels'].shape, ())

            # But it should also be possible to pass a list of column names in to get a 1-dimensional
            # array as "graph_labels" instead (for multitask learning for example)
            molecule_net = MoleculeNetDataset(
                data_directory=path,
                file_name=csv_path,
                dataset_name='test'
            )
            molecule_net.prepare_data()
            molecule_net.read_in_memory(label_column_name=['label', 'name'])
            molecule_net.set_attributes()

            molecule = molecule_net[0]
            self.assertIsInstance(molecule['graph_labels'], np.ndarray)
            self.assertEqual(molecule['graph_labels'].shape, (2, ))




import unittest

import numpy as np

from kgcnn.graph.base import GraphDict


class TestGraphDict(unittest.TestCase):

    def test_creating_empty_one_basically_works(self):
        """
        If it is possible to create a GraphDict instance with empty constructor
        """
        g = GraphDict()
        self.assertIsInstance(g, GraphDict)
        self.assertEqual(0, len(g))

    def test_creating_graph_dict_from_existing_dict_works(self):
        """
        If it is basically possible to create a GraphDict object by passing an already existing dict to
        the constructor
        """
        data = {
            'graph_labels': np.array([0.1]),
            'node_indices': np.array([0, 1, 2]),
            'node_attributes': np.array([[1], [1], [1]]),
            'edge_indices': np.array([
                [0, 1], [1, 0],
                [1, 2], [2, 1],
                [2, 0], [0, 2]
            ]),
            'edge_attributes': np.array([[1], [1], [1], [1], [1], [1]])
        }
        g = GraphDict(data)
        self.assertIsInstance(g, GraphDict)
        self.assertIsInstance(g['node_attributes'], np.ndarray)
        self.assertEqual((3, ), g['node_indices'].shape)
        self.assertEqual((3, 1), g['node_attributes'].shape)



import importlib
from typing import Union
from kgcnn.utils.serial import deserialize


def get_preprocessor(name: Union[str, dict], **kwargs):
    """Get a preprocessor.

    Args:
        name (str, dict): Serialization dictionary of class. This can also be a name of former graph functions for
            backward compatibility that now coincides with the processor's default name.
        kwargs: Kwargs for processor initialization, if :obj:`name` is string.

    Returns:
        GraphPreProcessorBase: Instance of graph preprocessor.
    """
    preprocessor_identifier = {
        "make_undirected_edges": "MakeUndirectedEdges",
        "add_edge_self_loops": "AddEdgeSelfLoops",
        "sort_edge_indices": "SortEdgeIndices",
        "set_edge_indices_reverse": "SetEdgeIndicesReverse",
        "pad_property": "PadProperty",
        "set_edge_weights_uniform": "SetEdgeWeightsUniform",
        "normalize_edge_weights_sym": "NormalizeEdgeWeightsSymmetric",
        "set_range_from_edges": "SetRangeFromEdges",
        "set_range": "SetRange",
        "set_angle": "SetAngle",
        "set_range_periodic": "SetRangePeriodic",
        "expand_distance_gaussian_basis": "ExpandDistanceGaussianBasis",
        "atomic_charge_representation": "AtomicChargesRepresentation",
        "principal_moments_of_inertia": "PrincipalMomentsOfInertia",
    }
    if isinstance(name, dict):
        return deserialize(name)
    # if given as string name. Lookup identifier.
    obj_class = getattr(importlib.import_module(str("kgcnn.graph.preprocessor")), str(preprocessor_identifier[name]))
    return obj_class(**kwargs)


def get_postprocessor(name: Union[str, dict], **kwargs):
    r"""Get a postprocessor.

    Args:
        name (str, dict): Serialization dictionary of class. This can also be a name of former graph functions for
            backward compatibility that now coincides with the processor's default name.
        kwargs: Kwargs for processor initialization, if :obj:`name` is string.

    Returns:
        GraphPostProcessorBase: Instance of graph postprocessor.
    """
    preprocessor_identifier = {
        "extensive_energy_force_scaler": "ExtensiveEnergyForceScalerPostprocessor",
    }
    if isinstance(name, dict):
        return deserialize(name)
    # if given as string. Lookup identifier.
    obj_class = getattr(importlib.import_module(str("kgcnn.graph.postprocessor")), str(preprocessor_identifier[name]))
    return obj_class(**kwargs)